import functools
import heapq
import pathlib
from typing import Dict, Tuple, Any, List, Iterable, Mapping
from .extensions.merge_methods import value_to_node
from .extensions.model_configs import ModelConfig
from .extensions import merge_methods
from .recipe_nodes import RecipeNode, RecipeNodeOrValue
from sd_mecha.merging import open_input_dicts, infer_model_configs


def convert(recipe: RecipeNodeOrValue, config: str | ModelConfig | RecipeNode, model_dirs: Iterable[pathlib.Path] = ()):
    """
    Convert a recipe from one model config to another.

    This searches for a chain of registered conversion functions that transform `recipe`’s underlying
    config into the target config, then composes them. For example, you might need to
    convert a LoRA adapter into the base model’s format.

    Args:
        recipe:
            A `RecipeNode` or dictionary representing the input model or partial recipe.
        config (str, ModelConfig or RecipeNode):
            The desired output config, or a recipe node that has the desired config.
        model_dirs (Iterable[Path], optional):
            Directories to resolve relative model paths.

    Returns:
        A new recipe node describing the entire conversion.

    Raises:
        ValueError:
            If no conversion path is found.
    """
    model_dirs = list(model_dirs)
    all_converters = merge_methods.get_all_converters()
    converter_paths: Dict[str, List[Tuple[str, Any]]] = {}
    for converter in all_converters:
        input_configs = converter.get_input_configs()
        return_config = converter.get_return_config(input_configs.args, input_configs.kwargs)
        src_config = input_configs.args[0].identifier
        tgt_config = return_config.identifier
        converter_paths.setdefault(src_config, [])
        converter_paths.setdefault(tgt_config, [])
        converter_paths[src_config].append((tgt_config, converter))

    if isinstance(config, RecipeNode):
        with open_input_dicts(config, model_dirs):
            config = config.model_config

    tgt_config = config if isinstance(config, str) else config.identifier

    if isinstance(recipe, Mapping):
        possible_configs = infer_model_configs(recipe)
        for possible_config in (cfg for s in possible_configs for cfg in s):
            res = create_conversion_recipe(recipe, converter_paths, possible_config.identifier, tgt_config)
            if res is not None:
                return res
        raise ValueError(
            "could not infer the intended config to convert from. "
            "explicitly specifying the input config might resolve the issue"
        )

    recipe = value_to_node(recipe)
    with open_input_dicts(recipe, model_dirs):
        src_config = recipe.model_config.identifier
    if src_config == "structural":
        raise ValueError(
            "recipe config is 'structural': "
            "structural recipes cannot be composed of any config conversions"
        )
    res = create_conversion_recipe(recipe, converter_paths, src_config, tgt_config)
    if res is None:
        raise ValueError(f"no config conversion exists from {src_config} to {tgt_config}")
    return res


def create_conversion_recipe(recipe, paths, src_config, tgt_config):
    shortest_path = dijkstra(paths, src_config, tgt_config)
    if shortest_path is None:
        return None
    return functools.reduce(lambda v, mm: mm(v), shortest_path, recipe)


def dijkstra(graph, start, goal):
    """
    graph: Dict[str, List[Tuple[str, any_id]]]
        For each node (str), a list of (neighbor_node, edge_id).
    start: str
    goal: str

    Returns: List of edge IDs (in order) that forms the shortest path from start to goal,
             or None if no path exists.
    """

    distances = {node: float('inf') for node in graph}
    distances[start] = 0
    predecessors = {node: None for node in graph}  # will store the node we came from
    edge_used = {node: None for node in graph}  # will store which edge ID led here
    heap = [(0, start)]

    while heap:
        current_dist, current_node = heapq.heappop(heap)
        if current_dist > distances[current_node]:
            continue

        if current_node == goal:
            break

        for neighbor, edge_id in graph[current_node]:
            distance_via_current = current_dist + 1
            if distance_via_current < distances[neighbor]:
                distances[neighbor] = distance_via_current
                predecessors[neighbor] = current_node
                edge_used[neighbor] = edge_id
                heapq.heappush(heap, (distance_via_current, neighbor))

    if distances[goal] == float('inf'):
        return None

    path_ids = []
    node = goal
    while node != start:
        path_ids.append(edge_used[node])
        node = predecessors[node]

    path_ids.reverse()
    return path_ids



from .merging import merge, open_input_dicts, infer_model_configs
from .serialization import serialize, deserialize, deserialize_path
from .streaming import StateDictKeyError
from .extensions.merge_methods import merge_method, value_to_node, RecipeNodeOrValue, Parameter, Return, StateDict
from .conversion import convert
from sd_mecha.extensions.builtin.merge_methods import (
    weighted_sum,
    slerp,
    n_average,
    geometric_median,
    subtract,
    perpendicular_component,
    geometric_sum,
    train_difference_mask,
    add_opposite_mask,
    add_strict_opposite_mask,
    add_cosine_a,
    add_cosine_b,
    ties_sum,
    ties_sum_extended,
    crossover,
    clamp,
    model_stock,
    fallback,
    cast,
    get_dtype,
    get_device,
    pick_component,
    omit_component,
    exchange_ema,
)
from .merge_method_wrappers import (
    add_difference,
    add_perpendicular,
    add_difference_ties,
    add_difference_ties_extended,
    copy_region,
    tensor_sum,
    rotate,
    dropout,
    ties_with_dare,
    n_model_stock,
)
from .helpers import model, literal, Defaults, set_log_level
from . import recipe_nodes, extensions


def _load_builtin_extensions():
    import sd_mecha.extensions.builtin.model_configs
    import sd_mecha.extensions.builtin.lycoris


_load_builtin_extensions()



import re
from typing import TypeVar
from sd_mecha.extensions.merge_methods import Return, Parameter, StateDict, merge_method


re_inp = re.compile(r"\.input_blocks\.(\d+)\.")  # 12
re_mid = re.compile(r"\.middle_block\.(\d+)\.")  # 1
re_out = re.compile(r"\.output_blocks\.(\d+)\.")  # 12


T = TypeVar("T")


# srcs, in order of priority when disagreements occur:
# - https://github.com/hako-mikan/sd-webui-supermerger/blob/f14b3e5d0be9c510d199cca502c4148160f901bb/scripts/mergers/mergers.py#L1376
# - https://github.com/s1dlx/meh/blob/04af2c8d63744fb6c02d35d328a2c84380cca444/sd_meh/merge.py#L360
# - https://github.com/vladmandic/automatic/blob/e22d0789bddd3894364b0d59a4c9b3e456e89079/modules/merging/merge_utils.py#L64
@merge_method(is_conversion=True)
def convert_sdxl_blocks_to_sgm(
    blocks: Parameter(StateDict[T], model_config="sdxl-supermerger_blocks"),
    **kwargs,
) -> Return(T, model_config="sdxl-sgm"):
    sgm_key = kwargs["key"]

    block_key = "BASE"
    if sgm_key.startswith("model.diffusion_model."):
        block_key = "OUT08"
        if ".time_embed" in sgm_key or ".label_emb" in sgm_key:
            block_key = "BASE"  # before input blocks
        elif ".out." in sgm_key:
            block_key = "OUT08"  # after output blocks
        elif m := re_inp.search(sgm_key):
            block_key = f"IN{int(m.groups(1)[0]):02}"
        elif re_mid.search(sgm_key):
            block_key = "M00"
        elif m := re_out.search(sgm_key):
            block_key = f"OUT{int(m.groups(1)[0]):02}"
    elif sgm_key.startswith("first_stage_model."):
        block_key = "VAE"

    return blocks[block_key]



import dataclasses
import torch
from typing import Iterable, Mapping, Dict
from sd_mecha.extensions import model_configs
from sd_mecha.extensions.merge_methods import merge_method, StateDict, Parameter, Return
from sd_mecha.extensions.model_configs import StateDictKey, ModelConfig, ModelConfigImpl, LazyModelConfigBase, KeyMetadata
from sd_mecha.streaming import StateDictKeyError


def _register_all_lycoris_configs():
    base_configs = model_configs.get_all_base()
    for base_config in base_configs:
        for lyco_config in (
                LycorisModelConfig(base_config, "lycoris", "lycoris", list(lycoris_algorithms)),
                LycorisModelConfig(base_config, "kohya", "lora", list(lycoris_algorithms)),
        ):
            model_configs.register_aux(lyco_config)
            lora_config_id = lyco_config.identifier
            base_config_id = lyco_config.base_config.identifier

            @merge_method(identifier=f"convert_'{lora_config_id}'_to_base", is_conversion=True)
            def diffusers_lora_to_base(
                lora: Parameter(StateDict[torch.Tensor], "weight", lora_config_id),
                **kwargs,
            ) -> Return(torch.Tensor, "delta", base_config_id):
                key = kwargs["key"]
                lycoris_keys = lyco_config.to_lycoris_keys(key)
                if not lycoris_keys:
                    raise StateDictKeyError(key)

                lycoris_key = next(iter(lycoris_keys))
                return compose_lora_up_down(lora, lycoris_key.split(".")[0])


def compose_lora_up_down(state_dict: Mapping[str, torch.Tensor], key: str):
    # fetching these 3 keys in lexicographic order is very important
    # any other order would raise the number of cache misses in the input safetensors when streaming keys
    #  which in turn would slow down merging significantly
    alpha = state_dict[f"{key}.alpha"]
    down_weight = state_dict[f"{key}.lora_down.weight"]
    up_weight = state_dict[f"{key}.lora_up.weight"]
    dim = down_weight.size(0)

    if up_weight.numel() <= down_weight.numel():
        up_weight = up_weight * (alpha / dim)
    else:
        down_weight = down_weight * (alpha / dim)

    if len(down_weight.size()) == 2:  # linear
        res = up_weight @ down_weight
    elif down_weight.size()[2:4] == (1, 1):  # conv2d 1x1
        res = (up_weight.squeeze(3).squeeze(2) @ down_weight.squeeze(3).squeeze(2)).unsqueeze(2).unsqueeze(3)
    else:  # conv2d 3x3
        res = torch.nn.functional.conv2d(down_weight.permute(1, 0, 2, 3), up_weight).permute(1, 0, 2, 3)
    return res


class LycorisModelConfig(LazyModelConfigBase):
    def __init__(self, base_config: ModelConfig, lycoris_identifier: str, prefix: str, algorithms: Iterable[str] | str):
        super().__init__()
        self.base_config = base_config
        self.lycoris_identifier = lycoris_identifier
        self.prefix = prefix
        self.algorithms = list(sorted(algorithms)) if not isinstance(algorithms, str) else [algorithms]

    @property
    def identifier(self) -> str:
        return f"{self.base_config.identifier}_{self.lycoris_identifier}_{'_'.join(self.algorithms)}"

    def create_config(self):
        if "lora" not in self.algorithms or len(self.algorithms) != 1:
            raise ValueError(f"unknown lycoris algorithms {self.algorithms}")

        identifier = f"{self.base_config.identifier}_{self.lycoris_identifier}_{'_'.join(self.algorithms)}"
        components = {
            k: _to_lycoris_keys(component.keys(), self.algorithms, self.prefix)
            for k, component in self.base_config.components().items()
        }
        return ModelConfigImpl(identifier, components)

    def to_lycoris_keys(self, key: StateDictKey) -> Mapping[StateDictKey, KeyMetadata]:
        return _to_lycoris_keys({key: KeyMetadata(None, None)}, self.algorithms, self.prefix)


def _to_lycoris_keys(
    keys: Mapping[StateDictKey, KeyMetadata],
    algorithms: Iterable[str],
    prefix: str,
) -> Dict[StateDictKey, KeyMetadata]:
    lycoris_keys = {}

    for algorithm in algorithms:
        for key, meta in keys.items():
            if key.endswith("bias") or not getattr(meta.metadata().dtype, "is_floating_point", True):
                continue

            key = key.split('.')
            if key[-1] == "weight":
                key = key[:-1]
            key = "_".join(key)

            for suffix in lycoris_algorithms[algorithm]:
                lycoris_key = f"{prefix}_{key}.{suffix}"
                lycoris_keys[lycoris_key] = dataclasses.replace(meta, shape=[], optional=True)

    return lycoris_keys


lycoris_algorithms = {
    "lora": ("lora_up.weight", "lora_down.weight", "alpha"),
}


_register_all_lycoris_configs()



import torch
from typing import Optional, Dict
from torch import Tensor
from .extensions.merge_methods import value_to_node, RecipeNodeOrValue, Parameter
from . import recipe_nodes
from sd_mecha.extensions.builtin import merge_methods
from sd_mecha.extensions.builtin.merge_methods import (
    subtract,
    ties_sum,
    ties_sum_extended,
    clamp,
    model_stock,
)


def add_difference(
    a: RecipeNodeOrValue, b: RecipeNodeOrValue, c: Optional[RecipeNodeOrValue] = None, *,
    alpha: Parameter(Tensor) = 1.0,
    clamp_to_ab: Parameter(bool) = False,
) -> recipe_nodes.RecipeNode:
    a = value_to_node(a)
    b = value_to_node(b)
    original_b = b

    if c is not None:
        c = value_to_node(c)
        b = subtract(
            b, c,
        )

    res = merge_methods.add_difference(
        a, b,
        alpha=alpha,
    )

    if a.merge_space == original_b.merge_space:
        b = original_b

    if clamp_to_ab:
        if a.merge_space != b.merge_space:
            raise TypeError(f"Merge space of A {a.merge_space} and B {b.merge_space} must be the same to clamp the merge.")
        res = clamp(
            res, a, b,
        )

    return res


def add_perpendicular(
    a: RecipeNodeOrValue, b: RecipeNodeOrValue, c: RecipeNodeOrValue, *,
    alpha: Parameter(Tensor) = 1.0,
) -> recipe_nodes.RecipeNode:
    a = value_to_node(a)
    b = value_to_node(b)
    c = value_to_node(c)

    a_diff = subtract(
        a, c,
    )
    b_diff = subtract(
        b, c,
    )

    perp_diff = merge_methods.perpendicular_component(
        a_diff, b_diff,
    )

    return merge_methods.add_difference(
        a, perp_diff,
        alpha=alpha,
    )


# latex notes in reference to original implementation: https://arxiv.org/abs/2306.01708
# - `base`: $$ \theta_{init} $$
# - `*models`: $$ \{\theta_{init}\}_{t=1}^n $$
# - `models` after `subtract`: $$ \tau_t $$
# - `alpha`: $$ \lambda $$
# - `k`: $$ k $$ ( From $$ \% $$ to $$ 1 $$ )
# - `res`: $$ \lambda * \tau_m $$
# - `return`: $$ \theta_m $$
# Special mode "TIES-SOUP" has been implemented by setting `vote_sgn` > 0.0
# Special mode "TIES-STOCK" has been implemented by setting `apply_stock` > 0.0
def add_difference_ties(
    base: RecipeNodeOrValue,
    *models: RecipeNodeOrValue,
    alpha: Parameter(Tensor) = 1.0,
    k: Parameter(float) = 1.0,
) -> recipe_nodes.RecipeNode:
    # $$ \{\theta_{init}\}_{t=1}^n $$
    base = value_to_node(base)
    models = tuple(value_to_node(model) for model in models)

    # Create task vectors.
    # $$ \tau_t $$
    models = tuple(
        subtract(model, base)
        if model.merge_space == "weight" else
        model
        for model in models
    )

    # step 1 + step 2 + step 3
    res = ties_sum(
        *models,
        k=k,
    )

    # Obtain merged checkpoint

    # $$ \theta_{init} + \lambda * \tau_m $$
    return add_difference(
        base, res,
        alpha=alpha,
    )


def add_difference_ties_extended(
    base: RecipeNodeOrValue,
    *models: RecipeNodeOrValue,
    alpha: Parameter(Tensor) = 1.0,
    k: Parameter(float) = 0.2,
    vote_sgn: Parameter(bool) = False,
    apply_stock: Parameter(bool) = False,
    cos_eps: Parameter(float) = 1e-6,
    apply_median: Parameter(bool) = False,
    eps: Parameter(float) = 1e-6,
    maxiter: Parameter(int) = 100,
    ftol: Parameter(float) = 1e-20,
) -> recipe_nodes.RecipeNode:
    # $$ \{\theta_{init}\}_{t=1}^n $$
    base = value_to_node(base)
    models = tuple(value_to_node(model) for model in models)

    # Create task vectors.
    # $$ \tau_t $$
    models = tuple(
        subtract(model, base)
        if model.merge_space == "weight" else
        model
        for model in models
    )

    # step 1 + step 2 + step 3
    res = ties_sum_extended(
        *models,
        k=k,
        vote_sgn=vote_sgn,
        apply_stock=apply_stock,
        cos_eps=cos_eps,
        apply_median=apply_median,
        eps=eps,
        maxiter=maxiter,
        ftol=ftol,
    )

    # Obtain merged checkpoint

    # $$ \theta_{init} + \lambda * \tau_m $$
    return add_difference(
        base, res,
        alpha=alpha,
    )


def copy_region(
    a: RecipeNodeOrValue, b: RecipeNodeOrValue, c: Optional[RecipeNodeOrValue] = None, *,
    width: Parameter(float) = 1.0,
    offset: Parameter(float) = 0.0,
    top_k: Parameter(bool) = False,
) -> recipe_nodes.RecipeNode:
    a = value_to_node(a)
    b = value_to_node(b)

    if c is not None:
        c = value_to_node(c)

        a = subtract(
            a, c,
        )
        b = subtract(
            b, c,
        )

    copy_method = [merge_methods.tensor_sum, merge_methods.top_k_tensor_sum][int(top_k)]
    res = copy_method(
        a, b,
        width=width,
        offset=offset,
    )

    if c is not None:
        res = merge_methods.add_difference(
            c, res,
            alpha=1.0,
        )

    return res


tensor_sum = copy_region


def rotate(
    a: RecipeNodeOrValue, b: RecipeNodeOrValue, c: Optional[RecipeNodeOrValue] = None, *,
    alignment: Parameter(float) = 1.0,
    alpha: Parameter(float) = 0.0,
    cache: Optional[Dict[str, torch.Tensor]] = None,
) -> recipe_nodes.RecipeNode:
    a = value_to_node(a)
    b = value_to_node(b)

    if c is not None:
        c = value_to_node(c)

        a = subtract(
            a, c,
        )
        b = subtract(
            b, c,
        )

    res = merge_methods.rotate(
        a, b,
        alignment=alignment,
        alpha=alpha,
    ).set_cache(cache)

    if c is not None:
        res = merge_methods.add_difference(
            c, res,
            alpha=1.0,
        )

    return res


def dropout(
    a: RecipeNodeOrValue,
    *models: RecipeNodeOrValue,
    probability: Parameter(float) = 0.9,
    alpha: Parameter(Tensor) = 0.5,
    overlap: Parameter(float) = 1.0,
    overlap_emphasis: Parameter(float) = 0.0,
    seed: Parameter(int) = None,
) -> recipe_nodes.RecipeNode:
    deltas = [
        subtract(model, a)
        for model in models
    ]
    ba_delta = merge_methods.dropout(*deltas, probability=probability, overlap=overlap, overlap_skew=overlap_emphasis, seed=seed)
    return merge_methods.add_difference(a, ba_delta, alpha=alpha)


ties_sum_with_dropout = merge_methods.ties_sum_with_dropout


# latex notes in reference to original implementation: https://arxiv.org/abs/2311.03099
# Notice that this is "TIES Merging w/ DARE", which is "Prune > Merge (TIES) > Rescale"
# See https://slgero.medium.com/merge-large-language-models-29897aeb1d1a for details
# - `base`: $$ \theta_{PRE} $$
# - `*models`: $$ \theta_{SFT}^{t_k} $$
# - `deltas`: $$ \delta^t = \theta_{SFT}^{t} - \theta_{PRE} \in \mathbb{R}^d $$
# - `probability`: $$ p $$
# - `res`: $$ \hat{\delta}^t = \tilde{\delta}^t / (1-p) $$
# - `alpha`: $$ \lambda $$
# - `k`: $$ k $$ ( From $$ \% $$ to $$ 1 $$ ) in TIES paper
# - `return`: $$ \theta_M = \theta_{PRE} + \lambda \cdot \Sigma_{k=1}^{K} \tilde{\delta}^{t_k} $$
# Special mode "TIES-SOUP" has been implemented by setting `vote_sgn` > 0.0
def ties_with_dare(
    base: RecipeNodeOrValue,
    *models: RecipeNodeOrValue,
    probability: Parameter(float) = 0.9,
    rescale: Parameter(bool) = True,
    alpha: Parameter(Tensor) = 1.0,
    seed: Parameter(int) = None,
    k: Parameter(float) = 1.0,
    vote_sgn: Parameter(bool) = False,
    apply_stock: Parameter(bool) = False,
    cos_eps: Parameter(float) = 1e-6,
    apply_median: Parameter(bool) = False,
    eps: Parameter(float) = 1e-6,
    maxiter: Parameter(int) = 100,
    ftol: Parameter(float) = 1e-20,
) -> recipe_nodes.RecipeNode:
    # $$ \delta^t = \theta_{SFT}^{t} - \theta_{PRE} \in \mathbb{R}^d $$
    base = value_to_node(base)
    models = tuple(value_to_node(model) for model in models)
    deltas = tuple(
        subtract(model, base)
        if model.merge_space == "weight" else
        model
        for model in models
    )

    # $$ \tilde{\delta}^{t_k} $$
    res = ties_sum_with_dropout(
        *deltas,
        probability=probability,
        rescale=rescale,
        k=k,
        vote_sgn=vote_sgn,
        seed=seed,
        apply_stock=apply_stock,
        cos_eps=cos_eps,
        apply_median=apply_median,
        eps=eps,
        maxiter=maxiter,
        ftol=ftol,
    )

    # $$ \theta_M = \theta_{PRE} + \lambda \cdot \Sigma_{k=1}^{K} \tilde{\delta}^{t_k} $$
    return merge_methods.add_difference(base, res, alpha=alpha)


# Following mergekit's implementation of Model Stock (which official implementation doesn't exist)
# https://github.com/arcee-ai/mergekit/blob/main/mergekit/merge_methods/model_stock.py
def n_model_stock(
    base: RecipeNodeOrValue,
    *models: RecipeNodeOrValue,
    cos_eps: Parameter(float) = 1e-6,
) -> recipe_nodes.RecipeNode:
    base = value_to_node(base)
    models = tuple(value_to_node(model) for model in models)
    deltas = tuple(
        subtract(model, base)
        if model.merge_space == "weight" else
        model
        for model in models
    )

    # This is hacky: Both w_avg and w_h will be calculated there.
    # Notice that t and cos_theta is vector instead of single value.
    # Conceptually it could compatable with TIES, but algorithm should be rewritten.
    res = model_stock(
        *deltas,
        cos_eps=cos_eps,
    )

    return merge_methods.add_difference(base, res, alpha=1.0)



import logging
import pathlib
import torch
from .extensions.merge_spaces import MergeSpace
from .extensions.model_configs import ModelConfig
from .merging import merge
from .conversion import convert
from .recipe_nodes import ModelRecipeNode, LiteralRecipeNode, RecipeNode, RecipeNodeOrValue
from .extensions.merge_methods import NonDictLiteralValue
from typing import Optional, List, MutableMapping, Iterable, Mapping


def model(
    state_dict: str | pathlib.Path | Mapping[str, torch.Tensor],
    config: Optional[str | ModelConfig] = None,
    merge_space: str | MergeSpace = "weight",
) -> RecipeNode:
    """
    Create a recipe node representing a state dict.

    Args:
        state_dict:
            Path to a `.safetensors` file or an already loaded state dict.
        config:
            Model config or an identifier thereof.
        merge_space:
            The merge space in which the model is expected to be.

    Returns:
        ModelRecipeNode: A node that can be used in recipe graphs.
    """
    if merge_space is None:
        raise ValueError("merge space cannot be None")

    if isinstance(state_dict, Mapping):
        if state_dict and not isinstance(first_value := next(iter(state_dict.values())), torch.Tensor):
            raise ValueError(f"state dict must contain values of type Tensor, not {type(first_value)}")
        return LiteralRecipeNode(state_dict, model_config=config, merge_space=merge_space)
    if isinstance(state_dict, str):
        state_dict = pathlib.Path(state_dict)
    return ModelRecipeNode(state_dict, model_config=config, merge_space=merge_space)


def literal(
    value: NonDictLiteralValue | dict,
    config: Optional[ModelConfig | str] = None,
    merge_space: Optional[str | MergeSpace] = None,
) -> LiteralRecipeNode:
    """
    Create a recipe node wrapping tensors or some builtin python objects.

    This is used to access recipe node properties on python objects directly, i.e.:
    ```python
    sd_mecha.literal({...}) | 3
    ```
    There is typically no need to wrap inputs into recipe nodes manually as this function is implicitly applied whenever needed.

    Args:
        value:
            The literal value to wrap into a recipe node.
        config:
            Model config or an identifier thereof.
        merge_space:
            The merge space in which the literal is expected to be.

    Returns:
        LiteralRecipeNode: A recipe node representing the literal value.
    """
    return LiteralRecipeNode(value, model_config=config, merge_space=merge_space)


def set_log_level(level: str = "INFO"):
    logging.basicConfig(format="%(levelname)s: %(message)s", level=level)


class Defaults:
    """
    Convenience class for common recipe operations to reduce repetition in recipe scripts.
    """

    def __init__(
        self,
        model_dirs: pathlib.Path | str | Iterable[pathlib.Path | str] = ...,
        merge_device: Optional[str | torch.device] = ...,
        merge_dtype: Optional[torch.dtype] = ...,
        output_device: Optional[str | torch.device] = ...,
        output_dtype: Optional[torch.dtype] = ...,
        threads: Optional[int] = ...,
        total_buffer_size: int = ...,
        strict_weight_space: bool = ...,
        check_finite: bool = ...,
        omit_extra_keys: bool = ...,
        omit_ema: bool = ...,
        check_mandatory_keys: bool = ...,
        tqdm: type = ...,
    ):
        """
        Args:
            See documentation for `sd_mecha.merge` or `sd_mecha.conversion` for a description of each parameter.
        """
        self.__model_dirs = model_dirs
        self.__merge_device = merge_device
        self.__merge_dtype = merge_dtype
        self.__output_device = output_device
        self.__output_dtype = output_dtype
        self.__threads = threads
        self.__total_buffer_size = total_buffer_size
        self.__strict_weight_space = strict_weight_space
        self.__check_finite = check_finite
        self.__omit_extra_keys = omit_extra_keys
        self.__omit_ema = omit_ema
        self.__check_mandatory_keys = check_mandatory_keys
        self.__tqdm = tqdm

    def convert(
        self,
        recipe: RecipeNodeOrValue,
        config: str | ModelConfig | RecipeNode,
        model_dirs: pathlib.Path | str | Iterable[pathlib.Path | str] = ...,
    ):
        """
        Convert a recipe or model from one model config to another.

        See `sd_mecha.convert` for more information.
        """
        if model_dirs is ...:
            model_dirs = self.__model_dirs
        model_dirs = self._model_dirs_to_pathlib_list(model_dirs)
        return convert(recipe, config, model_dirs)

    def merge(
        self,
        recipe: RecipeNodeOrValue,
        *,
        fallback_model: Optional[RecipeNodeOrValue] = ...,
        merge_device: Optional[str | torch.device] = ...,
        merge_dtype: Optional[torch.dtype] = ...,
        output_device: Optional[str | torch.device] = ...,
        output_dtype: Optional[torch.dtype] = ...,
        threads: Optional[int] = ...,
        total_buffer_size: int = ...,
        model_dirs: pathlib.Path | str | Iterable[pathlib.Path | str] = ...,
        strict_weight_space: bool = ...,
        check_finite: bool = ...,
        omit_extra_keys: bool = ...,
        omit_ema: bool = ...,
        check_mandatory_keys: bool = ...,
        tqdm: type = ...,
        output: MutableMapping[str, torch.Tensor] | pathlib.Path | str = ...,
    ) -> Optional[MutableMapping[str, torch.Tensor]]:
        """
        Materialize a state dict from a recipe graph and optionally save it to a file.

        See `sd_mecha.merge` for more information.
        """
        if merge_device is ...:
            merge_device = self.__merge_device
        if merge_dtype is ...:
            merge_dtype = self.__merge_dtype
        if output_device is ...:
            output_device = self.__output_device
        if output_dtype is ...:
            output_dtype = self.__output_dtype
        if threads is ...:
            threads = self.__threads
        if total_buffer_size is ...:
            total_buffer_size = self.__total_buffer_size
        if model_dirs is ...:
            model_dirs = self.__model_dirs
        model_dirs = self._model_dirs_to_pathlib_list(model_dirs)
        if strict_weight_space is ...:
            strict_weight_space = self.__strict_weight_space
        if check_finite is ...:
            check_finite = self.__check_finite
        if omit_extra_keys is ...:
            omit_extra_keys = self.__omit_extra_keys
        if omit_ema is ...:
            omit_ema = self.__omit_ema
        if check_mandatory_keys is ...:
            check_mandatory_keys = self.__check_mandatory_keys

        return merge(
            recipe,
            fallback_model=fallback_model,
            merge_device=merge_device,
            merge_dtype=merge_dtype,
            output_device=output_device,
            output_dtype=output_dtype,
            threads=threads,
            total_buffer_size=total_buffer_size,
            model_dirs=model_dirs,
            strict_weight_space=strict_weight_space,
            check_finite=check_finite,
            omit_extra_keys=omit_extra_keys,
            omit_ema=omit_ema,
            check_mandatory_keys=check_mandatory_keys,
            tqdm=tqdm,
            output=output,
        )

    @staticmethod
    def _model_dirs_to_pathlib_list(models_dir):
        if models_dir is ...:
            models_dir = []
        if not isinstance(models_dir, List):
            models_dir = [models_dir]
        models_dir = list(models_dir)
        for i in range(len(models_dir)):
            if isinstance(models_dir[i], str):
                models_dir[i] = pathlib.Path(models_dir[i])
            if models_dir[i] is not None:
                models_dir[i] = models_dir[i].absolute()

        return models_dir



import contextlib
import dataclasses
import functools
import gc
import itertools
import logging
import operator
import os
import pathlib
import sys
import threading
import typing
import torch
from .extensions import model_configs, model_formats
from .extensions.merge_methods import MergeMethod, StateDict, T as MergeMethodT, value_to_node
from .extensions.merge_spaces import MergeSpace
from .extensions.model_configs import ModelConfig, StructuralModelConfig, KeyMetadata
from .recipe_nodes import RecipeVisitor, LiteralRecipeNode, RecipeNode, MergeRecipeNode, ModelRecipeNode, RecipeNodeOrValue, NonDictLiteralValue
from .streaming import OutSafetensorsDict, TensorMetadata, StateDictKeyError
from . import recipe_nodes, serialization
from collections import OrderedDict
from concurrent.futures import ThreadPoolExecutor, as_completed, Future
from contextlib import nullcontext
from tqdm import tqdm as tqdm_original
from types import SimpleNamespace
from typing import Optional, Mapping, MutableMapping, List, Set, Iterable, Tuple, Dict, TypeVar
from .typing_ import is_subclass


def merge(
    recipe: RecipeNodeOrValue,
    *,
    fallback_model: Optional[RecipeNodeOrValue] = ...,
    merge_device: Optional[str | torch.device] = ...,
    merge_dtype: Optional[torch.dtype] = ...,
    output_device: Optional[str | torch.device] = ...,
    output_dtype: Optional[torch.dtype] = ...,
    threads: Optional[int] = ...,
    total_buffer_size: int = ...,
    model_dirs: Iterable[pathlib.Path] = ...,
    strict_weight_space: bool = ...,
    check_finite: bool = ...,
    omit_extra_keys: bool = ...,
    omit_ema: bool = ...,
    check_mandatory_keys: bool = ...,
    tqdm: type = ...,
    output: Optional[MutableMapping[str, torch.Tensor]] | pathlib.Path | str = ...,
) -> Optional[MutableMapping[str, torch.Tensor]]:
    """
    Materialize a state dict from a recipe graph and optionally save it to a file.

    For each key of the target model config, execute all instructions of the recipe graph
    and store the result into a dictionary using the specified output strategy.

    Args:
        recipe:
            A `RecipeNode`, python literal, or dictionary describing how to merge or transform multiple models.
        fallback_model (optional):
            A secondary recipe or model to provide values for any keys missing from `recipe`.
        merge_device (optional):
            Device to load intermediate tensors onto while merging (e.g., "cpu" or "cuda").
        merge_dtype (optional):
            Torch dtype for intermediate merges (e.g., `torch.float32`, `torch.float64`).
        output_device (optional):
            Final output device (e.g., "cpu").
        output_dtype (optional):
            Final dtype for the saved model.
        threads (optional):
            Number of threads to spawn for parallel merges. Defaults to a reasonable guess.
        total_buffer_size (optional):
            Total byte size of the buffers for all safetensors state dicts (input and output).
        model_dirs (optional):
            One or more directories to search for model files if `recipe` references relative paths.
        strict_weight_space (optional):
            If True, verifies that merges occur in "weight" space. If False, merges can happen
            in other merge spaces (like "delta" or "param").
        check_finite (optional):
            If True, warns if any non-finite values appear in the output model.
        omit_extra_keys (optional):
            If True, warns about and removes unrecognized keys from the output model.
        omit_ema (optional):
            If True, omits ema keys from the output model. Defaults to omit_extra_keys.
        check_mandatory_keys (optional):
            If True and an input model is missing non-optional keys, raises RuntimeError.
        tqdm (optional):
            A custom progress-bar factory. By default, uses `tqdm.tqdm`.
        output (optional):
            Where to store the merged state dict. Can be a filesystem path (string or
            `Path`) ending with `.safetensors`, an in-memory dict-like object, or None.
            If it is None or omitted, an empty dict is created and returned when the merge completes.

    Returns:
        The in-memory dictionary if `output` is either a MutableMapping or None, and nothing if `output` is a file path.
    """
    if output is ...:
        output = None
    if fallback_model is ...:
        fallback_model = None
    if merge_device is ...:
        merge_device = "cpu"
    if merge_dtype is ...:
        merge_dtype = torch.float64
    if output_device is ...:
        output_device = "cpu"
    if output_dtype is ...:
        output_dtype = torch.float16
    if threads is ...:
        threads = None
    if total_buffer_size is ...:
        total_buffer_size = 2**28
    if model_dirs is ...:
        model_dirs = ()
    else:
        model_dirs = list(model_dirs)
    if strict_weight_space is ...:
        strict_weight_space = True
    if check_finite is ...:
        check_finite = True
    if omit_extra_keys is ...:
        omit_extra_keys = True
    if omit_ema is ...:
        omit_ema = omit_extra_keys
    if check_mandatory_keys is ...:
        check_mandatory_keys = False
    if tqdm is ...:
        tqdm = tqdm_original

    recipe = value_to_node(recipe)
    original_recipe = recipe
    if fallback_model is not None:
        recipe = recipe | fallback_model

    if merge_device is not None or merge_dtype is not None:
        recipe = recipe.accept(CastInputDicts(merge_device, merge_dtype))

    if output_device is not None or output_dtype is not None:
        recipe = recipe.to(device=output_device, dtype=output_dtype)

    if threads is not None and (threads < 0 or not isinstance(threads, int)):
        raise RuntimeError("threads should be a non-negative integer or None")

    total_files_open = (
        recipe.accept(recipe_nodes.ModelsCountVisitor()) +
        int(isinstance(output, (str, pathlib.Path)))
    )
    buffer_size_per_file = total_buffer_size // max(1, total_files_open)
    if threads is None:
        threads = min(max(total_files_open, 2), os.cpu_count(), 8)

    if threads == 0:
        thread_local_data = SimpleNamespace()
        executor = ThisThreadExecutor()
    else:
        thread_local_data = threading.local()
        executor = ThreadPoolExecutor(max_workers=threads)

    with open_input_dicts(recipe, model_dirs, buffer_size_per_file, omit_extra_keys, check_mandatory_keys):
        if strict_weight_space and recipe.merge_space != "weight":
            raise ValueError(f"recipe should be in 'weight' space, not '{recipe.merge_space.identifier}'")

        recipe_metadata = recipe.model_config.metadata()
        if not omit_extra_keys:
            recipe, recipe_metadata = copy_extra_keys(recipe)

        if omit_ema and "ema" in recipe.model_config.components():
            for key in recipe.model_config.components()["ema"].keys():
                if key in recipe_metadata:
                    del recipe_metadata[key]

        buffer_size_per_file_per_thread = buffer_size_per_file // max(1, threads)

        with (
            executor,
            tqdm(total=len(recipe_metadata), desc="Merging recipe") as progress,
            _get_output_dict(
                output,
                recipe_metadata,
                original_recipe,
                model_dirs,
                buffer_size_per_file_per_thread,
            ) as output_dict,
        ):
            fix_torch_threading()
            futures = []
            for key in recipe_metadata:
                fn = recipe.accept
                fn = _track_output(fn, output_dict, key, check_finite)
                fn = _track_progress(fn, key, recipe_metadata[key].shape, progress)
                fn = _wrap_thread_context(fn, thread_local_data)
                futures.append(executor.submit(fn, KeyMergeVisitor(key)))

            for future in as_completed(futures):
                if future.exception() is not None:
                    for future_to_cancel in futures:
                        future_to_cancel.cancel()
                    raise future.exception()
                future.result()

            if isinstance(output_dict, MutableMapping):
                return output_dict


def fix_torch_threading():
    if not torch.cuda.is_available():
        return

    # this greedy loads the torch.linalg module
    # avoids a hard error caused by threads>1 with some torch ops
    # see https://github.com/pytorch/pytorch/issues/90613
    torch.linalg.inv(torch.ones((1, 1), device="cuda"))


def copy_extra_keys(recipe: RecipeNode):
    config = recipe.model_config
    merge_space = recipe.merge_space
    metadata = config.metadata()
    aliases = {k_alias: k for k, k_aliases in config.aliases().items() for k_alias in k_aliases}

    forwardable_nodes_visitor = ForwardableNodesVisitor(config, merge_space)
    recipe.accept(forwardable_nodes_visitor)
    forwardable_nodes = forwardable_nodes_visitor.forwardable_nodes

    new_recipe = functools.reduce(operator.or_, [n[0] for n in forwardable_nodes], recipe)
    new_metadata = OrderedDict((aliases.get(k, k), v) for n in forwardable_nodes for k, v in n[1].items()) | metadata
    return new_recipe, new_metadata


@dataclasses.dataclass
class ForwardableNodesVisitor(RecipeVisitor):
    target_config: ModelConfig
    target_merge_space: MergeSpace
    forwardable_nodes: List[Tuple[RecipeNode, Mapping[str, TensorMetadata]]] = dataclasses.field(default_factory=list)

    def visit_literal(self, node: LiteralRecipeNode):
        if not isinstance(node.value, Mapping) or not node.value:
            return

        if isinstance(next(iter(node.value.values())), RecipeNode):
            for v in node.value.values():
                v.accept(self)
        elif (
            isinstance(next(iter(node.value.values())), torch.Tensor) and
            node.model_config == self.target_config and
            node.merge_space == self.target_merge_space and
            not any(node in n[0] for n in self.forwardable_nodes)
        ):
            metadata = OrderedDict(
                (k, TensorMetadata(v.shape, v.dtype))
                for k, v in node.value.items()
            )
            self.forwardable_nodes.append((node, metadata))

    def visit_model(self, node: ModelRecipeNode):
        if node.model_config == self.target_config and not any(node in n[0] for n in self.forwardable_nodes):
            self.forwardable_nodes.append((node, OrderedDict(node.state_dict.metadata())))

    def visit_merge(self, node: MergeRecipeNode):
        for arg in (*node.args, *node.kwargs.values()):
            arg.accept(self)


@contextlib.contextmanager
def _get_output_dict(
    output: Optional[MutableMapping[str, torch.Tensor]] | pathlib.Path | str,
    merged_header: Mapping[str, TensorMetadata],
    recipe: RecipeNode,
    model_dirs: Iterable[pathlib.Path],
    buffer_size_per_thread: int,
):
    if isinstance(output, (str, pathlib.Path)):
        if not isinstance(output, pathlib.Path):
            output = pathlib.Path(output)
        if not output.is_absolute():
            for model_dir in model_dirs:
                output = model_dir / output
                break
        logging.info(f"Saving to {output}")

        try:
            serialized_recipe = serialization.serialize(recipe)
        except TypeError:
            logging.warning("The recipe graph could not be serialized. The output state dict will not contain the recipe.")
            serialized_recipe = None
        streamed_output = OutSafetensorsDict(
            output,
            merged_header,
            serialized_recipe,
            buffer_size_per_thread,
        )
        try:
            yield streamed_output
        finally:
            streamed_output.close()
    else:
        if output is None:
            output = {}
        yield output


def _track_output(fn, output, key, check_finite: bool):
    @functools.wraps(fn)
    def track_output(*args, **kwargs):
        try:
            res = fn(*args, **kwargs)
            if check_finite and isinstance(res, torch.Tensor) and not res.isfinite().all():
                logging.warning(f"there are non finite values in key '{key}'")
            output[key] = res
        except StateDictKeyError as k:
            logging.debug(f"skipping key {k}")
    return track_output


def _track_progress(fn, key, key_shape, progress):
    @functools.wraps(fn)
    def track_progress(*args, **kwargs):
        progress.set_postfix({"key": key} | ({"shape": list(key_shape)} if key_shape is not None else {}))
        res = fn(*args, **kwargs)
        progress.update()
        return res
    return track_progress


def _wrap_thread_context(fn, ctx):
    @functools.wraps(fn)
    def thread_context(*args, **kwargs):
        if torch.cuda.is_available():
            if not hasattr(ctx, 'cuda_stream'):
                setattr(ctx, "cuda_stream", torch.cuda.Stream())
            with torch.cuda.stream(ctx.cuda_stream):
                return fn(*args, **kwargs)
        else:
            return fn(*args, **kwargs)

    return thread_context


class ThisThreadExecutor(nullcontext):
    def submit(self, fn, /, *args, **kwargs):
        result = Future()
        result.set_result(fn(*args, **kwargs))
        return result


@contextlib.contextmanager
def open_input_dicts(
    recipe: recipe_nodes.RecipeNode,
    model_dirs: Iterable[pathlib.Path] = (),
    buffer_size_per_dict: int = 0,
    omit_extra_keys: bool = True,
    check_mandatory_keys: bool = False,
    empty_cuda_cache: bool = False,
):
    try:
        recipe.accept(LoadInputDictsVisitor(list(model_dirs), buffer_size_per_dict, omit_extra_keys, check_mandatory_keys))
        yield recipe
    finally:
        recipe.accept(CloseInputDictsVisitor())
        if empty_cuda_cache:
            gc.collect()
            torch.cuda.empty_cache()


@dataclasses.dataclass
class LoadInputDictsVisitor(RecipeVisitor):
    model_dirs: Iterable[pathlib.Path]
    buffer_size_per_dict: int
    strip_extra_keys: bool
    check_mandatory_keys: bool
    dicts_cache: MutableMapping[str, Mapping[str, torch.Tensor]] = dataclasses.field(default_factory=dict)
    structural_metadata: MutableMapping[str, KeyMetadata] = dataclasses.field(default_factory=OrderedDict)
    param_config: Optional[ModelConfig] = None

    def visit_literal(self, node: LiteralRecipeNode):
        if not isinstance(node.value, Mapping):
            return

        metadata = {}
        for k, v in node.value.items():
            if isinstance(v, RecipeNode):
                v.accept(self)
            elif isinstance(v, torch.Tensor):
                metadata[k] = KeyMetadata(v.shape, v.dtype, optional=True)
            elif k not in metadata:
                metadata[k] = KeyMetadata(None, None, optional=True)

        node.model_config = self.__determine_model_config(metadata, node.model_config)
        check_model_config(node.value, node.model_config, self.strip_extra_keys, self.check_mandatory_keys, "<in-memory state dict>")

    def visit_model(self, node: recipe_nodes.ModelRecipeNode):
        node.state_dict, node_path = self.__load_dict(node)
        metadata = OrderedDict(node.state_dict.metadata())
        node.model_config = self.__determine_model_config(metadata, node.model_config)
        check_model_config(node.state_dict, node.model_config, self.strip_extra_keys, self.check_mandatory_keys, str(node.path))

    def visit_merge(self, node: recipe_nodes.MergeRecipeNode):
        input_configs = node.merge_method.get_input_configs().as_dict(len(node.args))
        for k, v in itertools.chain(enumerate(node.args), node.kwargs.items()):
            v.accept(dataclasses.replace(self, param_config=input_configs[k]))

    def __load_dict(self, node: recipe_nodes.ModelRecipeNode):
        if node.state_dict is not None:
            return node.state_dict, node.path

        path = node.path
        if not isinstance(path, pathlib.Path):
            path = pathlib.Path(path)
        if not path.is_absolute():
            for base_dir in self.model_dirs:
                path_attempt = base_dir / path
                if path_attempt.exists():
                    path = path_attempt
                    break

        cache_key = str(path.resolve())
        if cache_key not in self.dicts_cache:
            matching_formats = []
            for model_format in model_formats.get_all():
                if model_format.matches(path):
                    matching_formats.append(model_format)

            if len(matching_formats) > 1:
                raise RuntimeError(f"ambiguous format ({', '.join(f.identifier for f in matching_formats)}) for model {path}")
            if len(matching_formats) < 1:
                raise RuntimeError(f"no matching format found for model {path}")

            self.dicts_cache[cache_key] = matching_formats[0].get_read_dict(path, self.buffer_size_per_dict)

        return self.dicts_cache[cache_key], path

    def __determine_model_config(self, metadata: MutableMapping[str, KeyMetadata], config_hint: Optional[ModelConfig]) -> ModelConfig:
        config = config_hint
        if config is None or config == model_configs.INFER:
            inferred_model_configs = infer_model_configs(metadata)
            if any(self.param_config in cfgs for cfgs in inferred_model_configs):
                config = self.param_config
            elif inferred_model_configs and len(inferred_model_configs[0]) == 1:
                config = next(iter(inferred_model_configs[0]))
            elif config is not None:  # config is INFER, structural is not allowed
                raise RuntimeError("could not infer model config")

        if config is None:
            if not self.structural_metadata:
                self.structural_metadata.update(metadata)
            else:
                for k in list(self.structural_metadata):
                    if k not in metadata:
                        del self.structural_metadata[k]
                    elif (
                        self.structural_metadata[k].shape is None or
                        metadata[k].shape is not None and metadata[k].metadata().shape.numel() > self.structural_metadata[k].metadata().shape.numel()
                    ):
                        self.structural_metadata[k] = metadata[k]
            config = StructuralModelConfig(self.structural_metadata)

        return config


def is_config_stub(config: Optional[model_configs.ModelConfig]) -> bool:
    return getattr(config, "identifier", None) in (None, model_configs.INFER.identifier)


def infer_model_configs(state_dict: Iterable[str]) -> List[Set[ModelConfig]]:
    state_dict_set = set(state_dict)
    configs_affinity = {}

    for model_config in model_configs.get_all():
        config_keys = set(model_config.keys()) | set(alias for aliases in model_config.aliases().values() for alias in aliases)
        matched_keys = state_dict_set.intersection(config_keys)

        # heuristic: accept config only if we match more than 10% of the keys of the state dict
        if len(matched_keys) >= len(state_dict_set) * 0.1:
            configs_affinity[model_config] = len(matched_keys)
        # heuristic: break early if we match more than 90% of the keys of a config
        if len(matched_keys) == len(state_dict_set) and len(matched_keys) >= len(model_config.keys()) * 0.9:
            break

    best_configs = sorted((
        {c for c, n in configs_affinity.items() if n == affinity}
        for affinity in set(configs_affinity.values())
    ), key=lambda s: configs_affinity[next(iter(s))], reverse=True)
    return best_configs


def check_model_config(state_dict: Iterable[str], config: ModelConfig, strip_extra_keys: bool, check_mandatory_keys: bool, state_dict_origin: str):
    state_dict_keys = set(state_dict)

    if strip_extra_keys:
        config_keys = set(k_alias for k, v in config.keys().items() for k_alias in [k, *v.aliases])
        extra_keys = state_dict_keys - config_keys
        if extra_keys:
            logging.warning(f"Found extra keys in state dict {state_dict_origin}: {extra_keys}")

    if check_mandatory_keys:
        missing_keys = set()
        for k, v in config.keys().items():
            if v.optional:
                continue

            for k_alias in v.aliases:
                if k_alias in state_dict_keys:
                    k = k_alias
                    break

            if k not in state_dict_keys:
                missing_keys.add(k)

        if missing_keys:
            raise RuntimeError(f"State dict {state_dict_origin} is missing non-optional keys: {missing_keys}")


@dataclasses.dataclass
class CloseInputDictsVisitor(RecipeVisitor):
    def visit_literal(self, node: LiteralRecipeNode):
        if isinstance(node.value, Mapping):
            if not node.value or isinstance(next(iter(node.value.values())), RecipeNode):
                for v in node.value.values():
                    v.accept(self)

    def visit_model(self, node: recipe_nodes.ModelRecipeNode):
        if node.state_dict is not None:
            node.state_dict.close()
        node.state_dict = None

    def visit_merge(self, node: recipe_nodes.MergeRecipeNode):
        for model in (*node.args, *node.kwargs.values()):
            model.accept(self)


@dataclasses.dataclass
class KeyMergeVisitor(RecipeVisitor):
    key: str

    def visit_literal(self, node: LiteralRecipeNode):
        value = node.value
        if isinstance(node.value, Mapping):
            value = self.__visit_sd(node.value, node.model_config)
        if isinstance(value, NonDictLiteralValue):
            return value
        if isinstance(value, RecipeNode):
            return value.accept(self)
        raise TypeError(f"Unexpected literal node value of type {type(value)}")

    def visit_model(self, node: recipe_nodes.ModelRecipeNode) -> torch.Tensor:
        return self.__visit_sd(node.state_dict, node.model_config)

    def __visit_sd(self, state_dict, model_config):
        aliases = model_config.aliases().get(self.key, [])
        keys = [self.key, *aliases]

        for i, key in enumerate(keys):
            try:
                return state_dict[key]
            except KeyError as e:
                if i == len(keys) - 1:
                    raise StateDictKeyError(self.key) from e

    def visit_merge(self, node: recipe_nodes.MergeRecipeNode) -> torch.Tensor:
        merged_args, merged_kwargs = self.__visit_deeper_first(node.args, node.kwargs, node.merge_method)
        return node.merge_method.merge_key(
            merged_args,
            merged_kwargs,
            self.key,
            node.cache,
        )

    def __visit_deeper_first(
        self,
        node_args: Tuple[recipe_nodes.RecipeNode, ...],
        node_kwargs: Dict[str, recipe_nodes.RecipeNode],
        merge_method: MergeMethod,
    ):
        def depth_of_value(index) -> int:
            nodes = node_args if isinstance(index, int) else node_kwargs
            return nodes[index].accept(recipe_nodes.ModelDepthRecipeVisitor())

        error_holder = ErrorHolder()
        merged = {}
        input_types = merge_method.get_input_types().as_dict(len(node_args))
        indices = itertools.chain(range(len(node_args)), node_kwargs.keys())

        for index in sorted(indices, key=depth_of_value, reverse=True):
            nodes = node_args if isinstance(index, int) else node_kwargs
            if is_subclass(input_types[index], StateDict):
                expected_type = next(iter(typing.get_args(input_types[index]) or (MergeMethodT,)))
                merged[index] = error_holder.intercept(MergeNodeWrapperStateDict, nodes[index], expected_type, self)
            else:
                merged[index] = cast_node_value(error_holder.intercept(nodes[index].accept, self), input_types[index])

        merged_args = [merged.get(index) for index in range(len(node_args))]
        merged_kwargs = {k: v for k, v in merged.items() if not isinstance(k, int)}
        error_holder.try_raise()
        return merged_args, merged_kwargs


class ErrorHolder:
    def __init__(self):
        self.exc_info = None

    def intercept(self, fn, *args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except Exception:
            self.exc_info = sys.exc_info()
            return None

    def try_raise(self):
        if self.exc_info:
            raise self.get_error()

    def get_error(self):
        if self.exc_info:
            return self.exc_info[1].with_traceback(self.exc_info[2])
        else:
            return None


class MergeNodeWrapperStateDict(StateDict):
    def __init__(
        self,
        merge_node: recipe_nodes.MergeRecipeNode,
        expected_type: type,
        original_merge_visitor: KeyMergeVisitor,
    ):
        self.merge_node = merge_node
        self.expected_type = expected_type
        self.original_merge_visitor = original_merge_visitor

    def __getitem__(self, key):
        key_merger = dataclasses.replace(self.original_merge_visitor, key=key)
        return cast_node_value(self.merge_node.accept(key_merger), self.expected_type)

    def __len__(self):
        return len(self.compute_keys())

    def __iter__(self):
        return iter(self.keys())

    def keys(self) -> Iterable[str]:
        return self.compute_keys().keys()

    @property
    def model_config(self) -> ModelConfig:
        return self.merge_node.model_config

    def compute_keys(self):
        return self.model_config.keys()


def cast_node_value(value, expected_type):
    if value is None:
        return value

    try:
        if issubclass(typing.get_origin(expected_type) or expected_type, StateDict):
            expected_type = (typing.get_args(expected_type) + (MergeMethodT,))[0]
    except TypeError:
        pass

    if isinstance(expected_type, TypeVar) or isinstance(value, expected_type):
        return value
    if isinstance(expected_type, str):
        raise RuntimeError(f"cannot implicitly convert {type(value)} to {expected_type}")
    if issubclass(expected_type, int):
        return int(value)
    if issubclass(expected_type, float):
        return float(value)
    if issubclass(expected_type, torch.Tensor):
        return torch.tensor(value, dtype=torch.float32)
    return value


@dataclasses.dataclass
class CastInputDicts(RecipeVisitor):
    device: str | torch.device
    dtype: torch.dtype

    def visit_literal(self, node: LiteralRecipeNode):
        if (
            isinstance(node.value, torch.Tensor | int | float) or
            isinstance(node.value, Mapping) and isinstance(next(iter(node.value.values())), torch.Tensor | int | float)
        ):
            return node.to(device=self.device, dtype=self.dtype)
        if isinstance(node.value, Mapping) and isinstance(next(iter(node.value.values())), RecipeNode):
            return LiteralRecipeNode(
                {k: v.accept(self) for k, v in node.value.items()},
                model_config=node.model_config,
                merge_space=node.merge_space,
            )
        return node

    def visit_model(self, node: ModelRecipeNode):
        return node.to(device=self.device, dtype=self.dtype)

    def visit_merge(self, node: MergeRecipeNode):
        return MergeRecipeNode(
            node.merge_method,
            tuple(v.accept(self) for v in node.args),
            {k: v.accept(self) for k, v in node.kwargs.items()},
            node.cache,
        )



import abc
import dataclasses
import itertools
import fuzzywuzzy.process
import inspect
import pathlib
import torch
import typing
from sd_mecha.recipe_nodes import RecipeNode, ModelRecipeNode, MergeRecipeNode, LiteralRecipeNode, NonDictLiteralValue, RecipeNodeOrValue, PythonLiteralValue
from . import merge_spaces, model_configs
from .merge_spaces import MergeSpace, MergeSpaceSymbol, AnyMergeSpace
from .model_configs import ModelConfig
from types import SimpleNamespace
from typing import Optional, Callable, Dict, Tuple, List, Iterable, Any, Generic, TypeVar, Mapping
from ..typing_ import is_subclass


T = TypeVar('T', *typing.get_args(NonDictLiteralValue))


class StateDict(Mapping[str, T], Generic[T], abc.ABC):
    @property
    @abc.abstractmethod
    def model_config(self) -> ModelConfig:
        pass

    @abc.abstractmethod
    def keys(self) -> Iterable[str]:
        pass


@dataclasses.dataclass
class ParameterData:
    interface: type | TypeVar
    merge_space: Optional[MergeSpace | AnyMergeSpace]
    model_config: Optional[ModelConfig]


@dataclasses.dataclass
class ParameterType:
    data: ParameterData


def Parameter(
    interface: type[NonDictLiteralValue | StateDict[NonDictLiteralValue]] | TypeVar,
    merge_space: Optional[MergeSpace | str | Iterable[MergeSpace | str] | MergeSpaceSymbol] = None,
    model_config: Optional[ModelConfig | str] = None,
) -> type[Any]:
    """
    Describe a parameter to a merge method with its type, optional merge space, and optional model config.

    `Parameter` is used in function type hints to specify how `@merge_method` should handle each
    argument. For example, `Parameter(Tensor, "weight")` means "this argument is a tensor
    in weight space."

    Args:
        interface (type):
            The Python or Torch type of this parameter.
        merge_space (str or Iterable[str], optional):
            Which merge space(s) are valid for this parameter (e.g., "weight", "delta").
        model_config (str or ModelConfig, optional):
            A specific model config or config identifier if it must match a certain architecture.

    Returns:
        A special type annotation object used by `@merge_method` to interpret function arguments.
    """
    supported_types = [StateDict] + list(T.__constraints__)
    if type(None) in (typing.get_args(interface) or ()):
        interface = typing.get_args(interface)[0]

    if not isinstance(interface, TypeVar) and not any(issubclass(typing.get_origin(interface) or interface, supported_type) for supported_type in supported_types):
        raise TypeError(f"type {interface} should be one of {', '.join(map(lambda x: x.__name__, supported_types))}")

    if isinstance(merge_space, (str, MergeSpace)):
        merge_space = (merge_space,)
    if isinstance(merge_space, Iterable):
        merge_space = {
            merge_spaces.resolve(m) if isinstance(m, str) else m
            for m in merge_space
        }

    if isinstance(model_config, str):
        model_config = model_configs.resolve(model_config)

    if getattr(model_config, "identifier", None) == "structural":
        raise ValueError("merge methods cannot convert 'structural' model configs")

    return type(Parameter.__name__, (ParameterType,), {
        "data": ParameterData(interface, merge_space, model_config)
    })


@dataclasses.dataclass
class ReturnType:
    data: ParameterData


def Return(
    interface: type[NonDictLiteralValue] | TypeVar,
    merge_space: Optional[MergeSpace | str | MergeSpaceSymbol] = None,
    model_config: Optional[ModelConfig | str] = None,
) -> type[Any]:
    """
    Describe the return type of a merge method, optionally including its merge space and model config.

    Args:
        interface (type):
            The Python or Torch type (e.g., `torch.Tensor`) returned by the merge method.
        merge_space (MergeSpace, str or MergeSpaceSymbol, optional):
            The single merge space valid for the return, or a symbol that depends on the input spaces.
        model_config (ModelConfig or str, optional):
            The model config that the returned tensor or dictionary should belong to.

    Returns:
        A type annotation object used by `@merge_method` for the return signature.
    """
    if not isinstance(interface, TypeVar):
        supported_types = list(T.__constraints__)
        if not any(issubclass(typing.get_origin(interface) or interface, supported_type) for supported_type in supported_types):
            raise TypeError(f"type {interface} should be one of {', '.join(map(str, supported_types))}")

    if isinstance(merge_space, str):
        merge_space = merge_spaces.resolve(merge_space)

    if isinstance(model_config, str):
        model_config = model_configs.resolve(model_config)

    if getattr(model_config, "identifier", None) == "structural":
        raise ValueError("merge methods cannot convert 'structural' model configs")

    return type(Return.__name__, (ReturnType,), {
        "data": ParameterData(interface, merge_space, model_config)
    })


P = TypeVar('P')
@dataclasses.dataclass
class FunctionArgs(Generic[P]):
    args: List[P]
    vararg: P | SimpleNamespace  # using SimpleNamespace as "empty" because P can be Optional
    kwargs: Dict[str, P]

    def as_dict(self, args_count=None) -> Dict[int | str, P]:
        varargs_count = self._get_varargs_count(args_count)
        args_dict = {i: v for i, v in enumerate(self.args)}
        vararg_dict = {
            i: self.vararg
            for i in range(len(args_dict), len(args_dict) + varargs_count)
        }
        return args_dict | vararg_dict | self.kwargs

    def args_varargs(self, args_count=None) -> List[P]:
        varargs_count = self._get_varargs_count(args_count)
        varargs = [self.vararg]*varargs_count
        return self.args + varargs

    def _get_varargs_count(self, args_count):
        n_args = len(self.args)
        if args_count is None:
            args_count = n_args + 1

        return max(0, args_count - n_args) * int(self.has_varargs())

    def has_varargs(self):
        return self.vararg != FunctionArgs.EMPTY_VARARGS


FunctionArgs.EMPTY_VARARGS = SimpleNamespace()


class MergeMethod:
    def __init__(self, fn: Callable, identifier: str):
        self.__wrapped__ = fn
        self.__f_spec = inspect.getfullargspec(self.__wrapped__)
        self.__f_hints = typing.get_type_hints(self.__wrapped__)
        self.identifier = identifier
        self.has_varkwargs = True
        self.default_merge_space = MergeSpaceSymbol(*merge_spaces.get_all())
        self.__validate_f()

    def __validate_f(self):
        names = self.get_param_names()
        params = self.get_params()  # validates param type annotations
        defaults = self.get_default_args()
        input_merge_spaces = self.get_input_merge_spaces()
        input_configs = self.get_input_configs()

        if self.__f_spec.varkw is None:
            self.has_varkwargs = False

        for param_idx in params.as_dict():
            is_default_arg = (
                isinstance(param_idx, int) and
                len(params.args) - len(defaults.args) <= param_idx < len(params.args)
            )
            is_default_kwarg = isinstance(param_idx, str) and param_idx in (self.__f_spec.kwonlydefaults or {})
            param_name = names.as_dict()[param_idx]
            if is_default_arg or is_default_kwarg:
                param_merge_space = input_merge_spaces.as_dict()[param_idx]
                if param_merge_space != {merge_spaces.resolve("param")}:
                    raise TypeError(f"The merge space for '{param_name}' should be 'param' since it has a default value.")

        input_configs_are_explicit = all(config is not None for config in input_configs.as_dict().values())
        if input_configs_are_explicit and self.get_return_config(input_configs.args_varargs(), input_configs.kwargs) is None:
            raise TypeError("Cannot infer the model config to return from the input model configs")

        return_data = self.__get_return_data(self.__f_hints.get("return"))  # validates return type annotation
        if isinstance(return_data.merge_space, MergeSpaceSymbol):
            if not any(k.merge_space for k in params.as_dict().values()):
                raise RuntimeError("when using a merge space symbol as output, it must also be used by at least one input parameter")

    def __repr__(self):
        return f"<merge method '{self.identifier}'>"

    def merge_key(
        self,
        input_args: Tuple[torch.Tensor | StateDict, ...],
        input_kwargs: Dict[str, torch.Tensor | StateDict],
        key: str,
        cache: Optional[dict],
    ):
        args, kwargs = self.__get_args_kwargs(input_args, input_kwargs, key, cache)
        return self.__wrapped__(*args, **kwargs)

    def __get_args_kwargs(
        self,
        input_args: Tuple[Any, ...],
        input_kwargs: Dict[str, float],
        key: str,
        cache: Optional[dict],
    ) -> Tuple[Tuple[torch.Tensor, ...], Dict]:
        if self.has_varkwargs:
            input_kwargs |= {
                "key": key,
                "cache": cache,
            }
        return input_args, input_kwargs

    def __call__(self, *args, **kwargs):
        return self.create_recipe(*args, **kwargs)

    def create_recipe(self, *args, **kwargs):
        params = self.get_param_names()
        defaults = self.get_default_args()
        first_default_arg = len(params.args) - len(defaults.args)

        first_arg_as_kwarg = min((
            *(params.args.index(k) for k in kwargs if k in params.args),
            float('inf'),
        ))

        def ensure_positive(v: int):
            if v < 0:
                raise RuntimeError
            return v

        max_args = len(params.args) if not params.has_varargs() else float("inf")
        min_args = len(params.args) - len(defaults.args)
        n_args = len(args) + len([kwargs[k] for k in params.args if k in kwargs])
        if not (min_args <= n_args <= max_args):
            raise TypeError(f"Expected from {min_args} to {max_args} arguments, received {n_args} arguments")

        args = [
            args[i] if i < min(first_arg_as_kwarg, len(args))
            else kwargs.pop(params.args[i]) if params.args[i] in kwargs
            else defaults.args[ensure_positive(i - first_default_arg)]
            for i in range(len(params.args))
        ] + list(args[len(params.args):])

        for k in kwargs:
            if k not in params.kwargs:
                raise TypeError(f"Unexpected keyword-argument '{k}'")

        for k in params.kwargs:
            if k not in (*kwargs.keys(), *defaults.kwargs.keys()):
                raise TypeError(f"Missing keyword-argument '{k}'")

        input_types = self.get_input_types()
        args = tuple(
            value_to_node(arg, arg_type)
            for arg, k, arg_type in zip(args, params.args_varargs(n_args), input_types.args_varargs(n_args))
        )
        kwargs = {
            k: value_to_node(arg, input_types.kwargs[k])
            for k, arg in (defaults.kwargs | kwargs).items()
        }

        return MergeRecipeNode(self, args, kwargs)

    def get_input_types(self) -> FunctionArgs[type]:
        params = self.get_params()
        arg_types = [
            param.interface
            for param in params.args
        ]
        vararg_type = params.vararg.interface if params.has_varargs() else FunctionArgs.EMPTY_VARARGS
        kwarg_types = {
            k: param.interface
            for k, param in params.kwargs.items()
        }
        return FunctionArgs(arg_types, vararg_type, kwarg_types)

    def get_return_merge_space(self, merge_space_args: List[MergeSpace], merge_space_kwargs: Dict[str, MergeSpace]) -> MergeSpace:
        names = self.get_param_names()
        n_args = len(merge_space_args)

        for idx, merge_space_arg in (*zip(names.args_varargs(n_args), merge_space_args), *merge_space_kwargs.items()):
            if merge_space_arg is None:
                raise ValueError(f"merge space of parameter {idx} cannot be None")

        input_merge_spaces = self.get_input_merge_spaces().as_dict(n_args)

        resolved_input_spaces = {}
        arg_tuples = enumerate(merge_space_args)
        kwarg_tuples = ((k, v) for k, v in merge_space_kwargs.items())
        for idx, merge_space_arg in itertools.chain(arg_tuples, kwarg_tuples):
            name = names.args_varargs(n_args)[idx] if isinstance(idx, int) else idx
            merge_space_param = input_merge_spaces[idx]
            is_symbol = isinstance(merge_space_param, MergeSpaceSymbol)
            valid_merge_spaces = merge_space_param.merge_spaces if is_symbol else merge_space_param
            if merge_space_arg not in valid_merge_spaces:
                valid_str_merge_spaces = tuple(m.identifier for m in valid_merge_spaces)
                raise TypeError(f"parameter '{name}' of method {self.identifier} expects a merge space in {valid_str_merge_spaces} but got {merge_space_arg.identifier}")
            if not is_symbol:
                continue

            if (resolved_input_space := resolved_input_spaces.get(merge_space_param)) is not None:
                # occurrence of already seen type var
                if merge_space_arg != resolved_input_space:
                    raise TypeError(f"parameter '{name}' of method {self.identifier} was resolved to {resolved_input_space.identifier} but got {merge_space_arg.identifier}")
            else:
                resolved_input_spaces[merge_space_param] = merge_space_arg

        merge_space_param: MergeSpace | MergeSpaceSymbol = self.__get_return_data(self.__f_hints.get("return")).merge_space
        if isinstance(merge_space_param, MergeSpaceSymbol):
            return resolved_input_spaces[merge_space_param]
        if merge_space_param is None:
            if self.default_merge_space in resolved_input_spaces:
                return resolved_input_spaces[self.default_merge_space]
            any_input_merge_space = next(iter(input_merge_spaces.values()))
            if all(v == any_input_merge_space for v in input_merge_spaces.values()):
                if isinstance(any_input_merge_space, set):
                    if len(any_input_merge_space) != 1:
                        raise RuntimeError(f"could not infer merge space of method '{self.identifier}'")
                    any_input_merge_space = next(iter(any_input_merge_space))
                return any_input_merge_space
            raise RuntimeError(f"could not infer merge space of method '{self.identifier}'")
        return merge_space_param

    def get_input_merge_spaces(self) -> FunctionArgs[AnyMergeSpace]:
        params = self.get_params()
        names = self.get_param_names()
        defaults = self.get_default_args()

        def get_merge_space_or_default(param, has_default):
            merge_space = param.merge_space
            if merge_space is None:
                if has_default or is_subclass(param.interface, PythonLiteralValue):
                    merge_space = {merge_spaces.resolve("param")}
                else:
                    merge_space = self.default_merge_space
            return merge_space

        args_merge_spaces = []
        for i in range(len(names.args)):
            param = params.args[i]
            merge_space = get_merge_space_or_default(param, i >= len(names.args) - len(defaults.args))
            args_merge_spaces.append(merge_space)

        varargs_merge_space = FunctionArgs.EMPTY_VARARGS
        if params.has_varargs():
            varargs_merge_space = params.vararg.merge_space
            if varargs_merge_space is None:
                varargs_merge_space = self.default_merge_space

        kwargs_merge_spaces = {}
        for name in names.kwargs:
            param = params.kwargs[name]
            merge_space = get_merge_space_or_default(param, name in defaults.kwargs)
            kwargs_merge_spaces[name] = merge_space

        return FunctionArgs(args_merge_spaces, varargs_merge_space, kwargs_merge_spaces)

    def get_return_config(self, arg_configs: List[Optional[ModelConfig]], kwarg_configs: Dict[str, Optional[ModelConfig]]) -> ModelConfig:
        input_configs = self.get_input_configs().as_dict(len(arg_configs))
        default_config = self.__get_return_data(self.__f_hints.get("return")).model_config

        arg_tuples = enumerate(arg_configs)
        kwarg_tuples = ((k, kwarg_configs.get(k)) for k in kwarg_configs)
        for param, arg_config in itertools.chain(arg_tuples, kwarg_tuples):
            if arg_config is None:
                continue

            param_config = input_configs[param]
            if param_config is None:
                param_config = default_config
            if param_config is None:
                param_config = arg_config
                default_config = arg_config
            if param_config.identifier != arg_config.identifier:
                raise ValueError(f"Recipe received an incompatible input: expected model config {param_config.identifier} but instead got {arg_config.identifier}")

        return default_config

    def get_input_configs(self) -> FunctionArgs[Optional[ModelConfig]]:
        params = self.get_params()
        return FunctionArgs(
            [arg.model_config for arg in params.args],
            params.vararg.model_config if params.has_varargs() else FunctionArgs.EMPTY_VARARGS,
            {k: v.model_config for k, v in params.kwargs.items()},
        )

    def get_param_names(self) -> FunctionArgs[str]:
        return FunctionArgs(
            self.__f_spec.args or [],
            self.__f_spec.varargs or FunctionArgs.EMPTY_VARARGS,
            {k: k for k in self.__f_spec.kwonlyargs} or {},
        )

    def get_default_args(self) -> FunctionArgs[Any]:
        return FunctionArgs(
            self.__f_spec.defaults or [],
            FunctionArgs.EMPTY_VARARGS,
            self.__f_spec.kwonlydefaults or {},
        )

    def get_params(self) -> FunctionArgs[ParameterData]:
        names = self.get_param_names()

        return FunctionArgs(
            [self.__get_parameter_data(self.__f_hints.get(k), k) for k in names.args],
            self.__get_parameter_data(self.__f_hints.get(names.vararg), names.vararg) if names.has_varargs() else FunctionArgs.EMPTY_VARARGS,
            {k: self.__get_parameter_data(self.__f_hints.get(k), k) for k in names.kwargs},
        )

    @staticmethod
    def __get_parameter_data(hint: type, param_name: str):
        hint_args = [arg for arg in (typing.get_args(hint) or ()) if arg is not type(None)]
        if hint_args:
            hint = hint_args[0]

        if hint is Parameter:
            raise TypeError(f"the type of parameter '{param_name}' should be `sd_mecha.Parameter(...)`, not `sd_mecha.Parameter` (note the lack of parentheses)")

        if not inspect.isclass(hint) or not issubclass(hint, ParameterType):
            raise TypeError(f"the type of parameter '{param_name}' should be 'sd_mecha.Parameter(...)', not '{getattr(hint, '__name__', hint)}'")
        return hint.data

    @staticmethod
    def __get_return_data(hint: type):
        if hint is Return:
            raise TypeError(f"the return type should be 'sd_mecha.Return(...)', not 'sd_mecha.Return' (note the lack of parentheses)")

        if not inspect.isclass(hint) or not issubclass(hint, ReturnType):
            raise TypeError(f"the return type should be 'sd_mecha.Return(...)', not '{getattr(hint, '__name__', hint)}'")
        return hint.data

    def get_identifier(self) -> str:
        return self.identifier


F = TypeVar("F", bound=Callable)


def merge_method(
    fn: Optional[F] = None, *,
    identifier: Optional[str] = None,
    register: bool = True,
    is_conversion: bool = False,
) -> MergeRecipeNode | Callable[[F], MergeRecipeNode]:
    """
    Decorator to define a custom merge method.

    This converts the decorated function into a `MergeMethod` object that can be used in
    recipe graphs. The type hints in the function signature determine which arguments are
    considered "weight" or "param" merges, and so on.

    Args:
        fn (callable, optional):
            The function to decorate. If omitted, the decorator can be used with named
            arguments (e.g. `@merge_method(is_conversion=True)`).
        identifier (str, optional):
            An explicit name to register this method under. By default, uses the function’s name.
        register (bool):
            If True (default), registers the merge method globally so it can be accessed by `merge_methods.resolve()`.
        is_conversion (bool):
            If True, marks this merge method as a config-conversion function. That means `convert()`
            will consider it as an implicit transition when converting between different model configs.

    Returns:
        A `MergeMethod` object or a decorator.
    """
    if fn is None:
        return lambda fn: __recipe_impl(fn, identifier=identifier, register=register, is_conversion=is_conversion)
    return __recipe_impl(fn, identifier=identifier, register=register, is_conversion=is_conversion)


def __recipe_impl(
    fn: Callable, *,
    identifier: Optional[str] = None,
    register: bool,
    is_conversion: bool,
):
    if identifier is None:
        identifier = fn.__name__
    fn_object = MergeMethod(fn, identifier)

    if register:
        _merge_methods_registry[identifier] = fn_object
        if is_conversion:
            _conversion_registry[identifier] = validate_config_conversion(fn_object)
    elif is_conversion:
        raise ValueError("A conversion recipe must be registered")

    return fn_object


def validate_config_conversion(merge_method: MergeMethod):
    params = merge_method.get_param_names()
    args_varargs = params.args if params.args else params.args_varargs()
    assert len(args_varargs) == 1, f"the merge method should be able to take exactly 1 positional argument"
    configs = merge_method.get_input_configs()
    input_config = configs.args if configs.args else configs.args_varargs()[0]
    assert input_config is not None, f"the input ModelConfig['identifier...'] is missing. It should be appended to the type annotation of `{args_varargs[0]}`"
    return merge_method


_merge_methods_registry = {}
_conversion_registry = {}


def resolve(identifier: str) -> MergeMethod:
    try:
        return _merge_methods_registry[identifier]
    except KeyError as e:
        suggestion = fuzzywuzzy.process.extractOne(str(e), _merge_methods_registry.keys())[0]
        raise ValueError(f"unknown merge method: {e}. Nearest match is '{suggestion}'")


def get_all() -> List[MergeMethod]:
    return list(_merge_methods_registry.values())


def get_all_converters() -> List[MergeMethod]:
    return list(_conversion_registry.values())


def value_to_node(node_or_value: RecipeNodeOrValue, expected_type: type = None) -> RecipeNode:
    """
    Convert a literal or path into a `RecipeNode`, if it isn't one already.

    This helper is primarily used internally when constructing recipe graphs. For instance,
    if you pass a float to a merge method where a tensor is expected, `value_to_node` wraps
    that float in a node that will automatically convert it into a tensor when broadcasting
    over state dict keys.

    Args:
        node_or_value:
            The input that should become a `RecipeNode`.
        expected_type (type, optional):
            The target type that the value should be converted to.

    Returns:
        A `RecipeNode` instance, if conversion is successful.
    """
    if isinstance(node_or_value, RecipeNode):
        return node_or_value

    if not isinstance(node_or_value, RecipeNodeOrValue):
        raise TypeError(f"type of 'node_or_value' should be one of {typing.get_args(RecipeNodeOrValue)}, not {type(node_or_value)}")

    if expected_type is None:
        if isinstance(node_or_value, Mapping):
            expected_type = next(iter(node_or_value.values()))
        else:
            expected_type = type(node_or_value)

    numeric = int | float

    # verify dict value type consistency
    if isinstance(node_or_value, Mapping) and node_or_value:
        actual_type = type(next(iter(node_or_value.values())))
        if is_subclass(actual_type, numeric):
            actual_type = numeric
        if is_subclass(actual_type, RecipeNode):
            actual_type = RecipeNode
        if not is_subclass(actual_type, NonDictLiteralValue | RecipeNode):
            raise TypeError(f"unsupported type found in input dict: {actual_type} (supported types are {typing.get_args(NonDictLiteralValue)})")
        if not all(isinstance(v, actual_type) for v in node_or_value.values()):
            bad_type = next(iter(type(v) for v in node_or_value.values() if not isinstance(v, actual_type)))
            raise TypeError(f"inconsistent types found in input dict: {actual_type} and {bad_type}")
        if is_subclass(actual_type, RecipeNode):
            first_node = next(iter(node_or_value.values()))
            if not all(r.model_config == first_node.model_config for r in node_or_value.values()):
                raise RuntimeError("inconsistent model configs found in recipe dict")
            if not all(r.merge_space == first_node.merge_space for r in node_or_value.values()):
                raise RuntimeError("inconsistent output merge space found in recipe dict")

    try:
        if issubclass(typing.get_origin(expected_type) or expected_type, StateDict):
            expected_type = next(iter(typing.get_args(expected_type) or (T,)))
    except TypeError:
        pass

    if isinstance(node_or_value, NonDictLiteralValue | Mapping):
        return LiteralRecipeNode(node_or_value)

    if isinstance(expected_type, TypeVar) and isinstance(node_or_value, pathlib.Path) or issubclass(expected_type, torch.Tensor):
        try:
            return ModelRecipeNode(node_or_value)
        except TypeError as e:
            base_error_message = f"No implicit conversion exists from {type(node_or_value)} to Dict[Tensor]"
            if isinstance(node_or_value, str):
                raise TypeError(f"{base_error_message}. Consider using pathlib.Path instead if a model path was intended") from e
            if not isinstance(node_or_value, pathlib.Path):
                raise TypeError(base_error_message) from e

    raise TypeError(f"No implicit conversion exists from {type(node_or_value)} to {expected_type}")



import pathlib
from sd_mecha.extensions import model_configs


yaml_directory = pathlib.Path(__file__).parent


def _register_configs():
    for yaml in yaml_directory.glob("*.yaml"):
        config = model_configs.YamlModelConfig(yaml)
        model_configs.register(config)


_register_configs()
from .convert_sdxl_kohya_to_original import convert_sdxl_kohya_to_original
from .convert_sdxl_diffusers_unet_to_original import convert_sdxl_diffusers_unet_to_original
from .convert_sd1_kohya_to_original import convert_sd1_kohya_to_original
from . import convert_sdxl_blocks
from . import convert_sd1_blocks



import functools
import math
import numpy as np
import torch
from scipy.stats import binom
from torch import Tensor
from typing import Tuple, TypeVar, Sequence
from .svd import orthogonal_procrustes, fractional_matrix_power, torch_svd_lowrank
from .ema import exchange_ema
from sd_mecha.extensions.merge_methods import merge_method, StateDict, Parameter, Return
from sd_mecha.streaming import StateDictKeyError


T = TypeVar("T")


@merge_method
def weighted_sum(
    a: Parameter(StateDict[Tensor]),
    b: Parameter(StateDict[Tensor]),
    alpha: Parameter(Tensor) = 0.5,
    **kwargs,
) -> Return(Tensor):
    key = kwargs["key"]

    if alpha.numel() == 1:
        alpha_float = alpha.item()
        if math.isclose(alpha_float, 0.0):
            return a[key]
        if math.isclose(alpha_float, 1.0):
            return b[key]

    return torch.lerp(a[key], b[key], alpha)


@merge_method
def n_average(
    *models: Parameter(Tensor),
) -> Return(Tensor):
    return torch.mean(torch.stack(models), dim=0)


@merge_method
def slerp(
    a: Parameter(Tensor),
    b: Parameter(Tensor),
    *,
    alpha: Parameter(Tensor) = 0.5,
) -> Return(Tensor):
    a_normalized = a / a.norm()
    b_normalized = b / b.norm()

    ab_dot = (a_normalized * b_normalized).sum().clamp(-1, 1)

    omega = torch.arccos(ab_dot)
    a_contrib = a_normalized * torch.sin((1-alpha)*omega)
    b_contrib = b_normalized * torch.sin(alpha*omega)
    res = (a_contrib + b_contrib) / torch.sin(omega)
    res *= torch.lerp(a.norm(), b.norm(), alpha)
    if res.isnan().any():
        return torch.lerp(a, b, alpha)
    return res


@merge_method
def add_difference(
    a: Parameter(StateDict[Tensor], "weight"),
    b: Parameter(StateDict[Tensor], "delta"),
    alpha: Parameter(Tensor) = 1.0,
    **kwargs,
) -> Return(Tensor, "weight"):
    key = kwargs["key"]
    if alpha.numel() == 1 and math.isclose(alpha.item(), 0.0):
        return a[key]

    b_val = b[key]  # try to load b from memory first in case it fails to merge before a
    return a[key].addcmul(b_val, alpha)


@merge_method
def subtract(
    a: Parameter(Tensor, "weight"),
    b: Parameter(Tensor, "weight"),
) -> Return(Tensor, "delta"):
    return a - b


@merge_method
def perpendicular_component(
    a: Parameter(Tensor),
    b: Parameter(Tensor),
) -> Return(Tensor):
    norm_a = torch.linalg.norm(a)
    res = b - a * (a / norm_a * (b / norm_a)).sum()
    if res.isnan().any():
        return torch.zeros_like(a)
    return res


@merge_method
def geometric_sum(
    a: Parameter(Tensor),
    b: Parameter(Tensor),
    alpha: Parameter(Tensor) = 0.5,
) -> Return(Tensor):
    a = torch.complex(a, torch.zeros_like(a))
    b = torch.complex(b, torch.zeros_like(b))
    res = a ** (1 - alpha) * b ** alpha
    return res.real


@merge_method
def add_cosine_a(
    a: Parameter(Tensor, "weight"),
    b: Parameter(Tensor, "weight"),
    alpha: Parameter(Tensor) = 1.0,
) -> Return(Tensor, "weight"):
    a_norm = torch.nn.functional.normalize(a, dim=0)
    b_norm = torch.nn.functional.normalize(b, dim=0)
    similarity = torch.nn.functional.cosine_similarity(a_norm, b_norm, dim=0)
    return add_cosine_generic(a, b, alpha, similarity)


@merge_method
def add_cosine_b(
    a: Parameter(Tensor, "weight"),
    b: Parameter(Tensor, "weight"),
    alpha: Parameter(Tensor) = 1.0,
) -> Return(Tensor, "weight"):
    similarity = torch.nn.functional.cosine_similarity(a, b, dim=0)
    dot_product = torch.sum(a * b)
    magnitude_similarity = dot_product / (torch.norm(a) * torch.norm(b))
    combined_similarity = (similarity + magnitude_similarity) / 2.0
    return add_cosine_generic(a, b, alpha, combined_similarity)


def add_cosine_generic(a: Tensor, b: Tensor, alpha: Tensor, similarity: Tensor) -> Tensor:
    k = 1 - torch.clamp(similarity - alpha, 0, 1)
    return torch.lerp(a, b, k)


@merge_method
def tensor_sum(
    a: Parameter(Tensor),
    b: Parameter(Tensor),
    width: Parameter(float) = 0.5,
    offset: Parameter(float) = 0.0,
) -> Return(Tensor):
    if a.shape == ():
        if width > 0.5:
            return b
        return a

    start_i, end_i, region_is_inverted = ratio_to_region(width, offset, a.size(0))
    if region_is_inverted:
        b[start_i:end_i] = a[start_i:end_i]
        return b
    else:
        a[start_i:end_i] = b[start_i:end_i]
        return a


@merge_method
def top_k_tensor_sum(
    a: Parameter(Tensor),
    b: Parameter(Tensor),
    width: Parameter(float) = 1.0,
    offset: Parameter(float) = 0.0,
) -> Return(Tensor):
    a_flat = torch.flatten(a)
    a_dist = torch.msort(a_flat)
    b_indices = torch.argsort(torch.flatten(b), stable=True)
    redist_indices = torch.argsort(b_indices)

    start_i, end_i, region_is_inverted = ratio_to_region(width, offset, torch.numel(a))
    start_top_k = kth_abs_value(a_dist, start_i)
    end_top_k = kth_abs_value(a_dist, end_i)

    indices_mask = (start_top_k <= torch.abs(a_dist)) & (torch.abs(a_dist) <= end_top_k)
    if region_is_inverted:
        indices_mask = ~indices_mask
    indices_mask = torch.gather(indices_mask.float(), 0, redist_indices)

    a_redist = torch.gather(a_dist, 0, redist_indices)
    a_redist = (1 - indices_mask) * a_flat + indices_mask * a_redist
    return a_redist.reshape_as(a)


def kth_abs_value(a: Tensor, k: int) -> Tensor:
    if k <= 0:
        return torch.tensor(-1, device=a.device)
    else:
        return torch.kthvalue(torch.abs(a.float()), k)[0]


def ratio_to_region(width: float, offset: float, n: int) -> Tuple[int, int, bool]:
    if width < 0:
        offset += width
        width = -width
    width = min(width, 1)

    if offset < 0:
        offset = 1 + offset - int(offset)
    offset = math.fmod(offset, 1.0)

    if width + offset <= 1:
        inverted = False
        start = offset * n
        end = (width + offset) * n
    else:
        inverted = True
        start = (width + offset - 1) * n
        end = offset * n

    return round(start), round(end), inverted


@merge_method
def train_difference_mask(
    a: Parameter(Tensor),
    b: Parameter(Tensor),
    c: Parameter(Tensor),
    alpha: Parameter(Tensor) = 1.0,
) -> Return(Tensor, "param"):
    return alpha * 1.8 * torch.nan_to_num((b - a).abs() / ((b - a).abs() + (b - c).abs()), nan=0)


@merge_method
def add_opposite_mask(
    a: Parameter(Tensor),
    b: Parameter(Tensor),
    c: Parameter(Tensor),
    alpha: Parameter(Tensor) = 1.0,
) -> Return(Tensor, "param"):
    return alpha * 2 * torch.nan_to_num((a - b).abs() / ((a - b).abs() + (a + b - 2*c).abs()), nan=0)


@merge_method
def add_strict_opposite_mask(
    a: Parameter(Tensor),
    b: Parameter(Tensor),
    c: Parameter(Tensor),
    alpha: Parameter(Tensor) = 1.0,
) -> Return(Tensor, "param"):
    threshold = torch.maximum(torch.abs(a - c), torch.abs(b - c))
    return alpha * torch.clamp(torch.nan_to_num((c - a) * (b - c) / threshold**2, nan=0), 0) * 2


@merge_method
def select_max_delta(
    a: Parameter(Tensor, "delta"),
    b: Parameter(Tensor, "delta"),
    alpha: Parameter(Tensor) = 0.5,
) -> Return(Tensor, "delta"):
    a_abs = ((a - a.mean()) / a.std()).nan_to_num(nan=0).abs()
    b_abs = ((b - b.mean()) / b.std()).nan_to_num(nan=0).abs()
    return torch.where((1 - alpha) * a_abs >= alpha * b_abs, a, b)


@merge_method
def select_max_white_delta(
    a: Parameter(Tensor, "delta"),
    b: Parameter(Tensor, "delta"),
    alpha: Parameter(Tensor) = 0.5,
) -> Return(Tensor, "delta"):
    a_norm = (a - a.mean()) / a.std(correction=0)
    b_norm = (b - b.mean()) / b.std(correction=0)

    d = torch.stack((a_norm.flatten(), b_norm.flatten()), dim=-1)
    v, vs = torch.linalg.eigh(d.cov(correction=0))
    w = vs @ torch.diag_embed(1 / v.sqrt()) @ vs.mH
    a_w, b_w = (w @ d).T

    return torch.where((1 - alpha) * a_w.abs() >= alpha * b_w.abs(), a, b)


@merge_method
def multiply_quotient(
    a: Parameter(Tensor),
    b: Parameter(Tensor),
    c: Parameter(Tensor),
    alpha: Parameter(Tensor) = 1.0,
) -> Return(Tensor):
    ac_log = torch.log(a.abs()) - torch.log(c.abs())
    bc_log = torch.log(b.abs()) - torch.log(c.abs())

    b = torch.complex(b, torch.zeros_like(b))
    c = torch.complex(c, torch.zeros_like(c))

    threshold = torch.maximum(torch.abs(ac_log), torch.abs(bc_log))
    alpha = alpha * torch.clamp(-torch.nan_to_num(ac_log * bc_log / threshold**2, nan=0), 0)

    res = a * (b / c)**alpha
    res = torch.where(torch.isnan(res), a, res)
    return res.real


@merge_method
def crossover(
    a: Parameter(Tensor),
    b: Parameter(Tensor),
    alpha: Parameter(float) = 0.5,
    tilt: Parameter(float) = 0.0,
) -> Return(Tensor):
    if alpha == 0:
        return a
    if alpha == 1:
        return b
    if tilt == 1:
        return torch.lerp(a, b, alpha)

    if len(a.shape) == 0 or torch.allclose(a.half(), b.half()):
        return torch.lerp(a, b, tilt)

    shape = a.shape

    a_dft = torch.fft.rfftn(a, s=shape)
    b_dft = torch.fft.rfftn(b, s=shape)

    dft_filter = create_filter(a_dft.shape, alpha, tilt, device=a.device)

    x_dft = (1 - dft_filter)*a_dft + dft_filter*b_dft
    return torch.fft.irfftn(x_dft, s=shape)


def create_filter(shape: Tuple[int, ...] | torch.Size, alpha: float, tilt: float, device=None):
    """
    Create a crossover filter. The cut is first tilted around (0, 0), then slid along its normal until it touches the point (alpha, 1 - alpha).
    :param shape: shape of the filter
    :param alpha: the ratio between the low frequencies and high frequencies. must be in [0, 1]
      0 = all 0s, 1 = all 1s, 0s correspond to low frequencies and 1s correspond to high frequencies
    :param tilt: tilt of the filter. 0 = vertical filter, 0.5 = 45 degrees, 1 = degenerates to a weighted sum with alpha=alpha
    :param device: device of the filter
    :return:
    """
    if not 0 <= alpha <= 1:
        raise ValueError("alpha must be between 0 and 1")

    # normalize tilt to the range [0, 4]
    tilt -= math.floor(tilt // 4 * 4)
    if tilt > 2:
        alpha = 1 - alpha
        alpha_inverted = True
    else:
        alpha_inverted = False

    gradients = list(reversed([
        torch.linspace(0, 1, s, device=device)
        if i == 0 or s == 1 else
        # negative frequencies are in the second half of the dimension
        torch.cat([
            torch.linspace(0, (s - 1) // 2, s - s // 2, device=device),
            torch.linspace(s // 2, 1, s // 2, device=device)
        ]) / (s // 2)
        for i, s in enumerate(reversed(shape))
    ]))

    if len(shape) > 1:
        grids = torch.meshgrid(*(g**2 for g in gradients), indexing='ij')
        mesh = (torch.stack(grids).sum(dim=0) / len(shape)).sqrt()
    else:
        mesh = gradients[0]

    if tilt < 1e-10 or abs(tilt - 4) < 1e-10:
        dft_filter = (mesh > 1 - alpha).float()
    elif abs(tilt - 2) < 1e-10:
        dft_filter = (mesh < 1 - alpha).float()
    else:
        tilt_cot = 1 / math.tan(math.pi * tilt / 2)
        if tilt <= 1 or 2 < tilt <= 3:
            dft_filter = mesh*tilt_cot + alpha*tilt_cot + alpha - tilt_cot
        else:  # 1 < tilt <= 2 or 3 < tilt
            dft_filter = mesh*tilt_cot - alpha*tilt_cot + alpha
        dft_filter = dft_filter.clip(0, 1)

    if alpha_inverted:
        dft_filter = 1 - dft_filter
    return dft_filter


@merge_method
def rotate(
    a_dict: Parameter(StateDict[Tensor]),
    b_dict: Parameter(StateDict[Tensor]),
    alignment: Parameter(float) = 1.0,
    alpha: Parameter(Tensor) = 0.0,
    **kwargs,
) -> Return(Tensor):
    key = kwargs["key"]

    if alpha.numel() == 1:
        alpha_float = alpha.item()
        if math.isclose(alignment, 0.0) and math.isclose(alpha_float, 0.0):
            return a_dict[key]
        if math.isclose(alignment, 1.0) and math.isclose(alpha_float, 1.0):
            return b_dict[key]

    a = a_dict[key]
    b = b_dict[key]
    if len(a.shape) <= 1 or torch.allclose(a.half(), b.half()):
        return (1-alpha)*a + alpha*b

    is_conv = len(a.shape) == 4 and a.shape[-2:].numel() != 1
    if is_conv:
        shape_2d = a.shape[:2].numel(), a.shape[2:].numel()
    else:
        shape_2d = a.shape[:1].numel(), a.shape[1:].numel()

    a_neurons = a.reshape(*shape_2d)
    b_neurons = b.reshape(*shape_2d)
    a_centroid = a_neurons.mean(0)
    b_centroid = b_neurons.mean(0)
    a_neurons -= a_centroid
    b_neurons -= b_centroid

    cache = kwargs.get("cache")
    if cache is not None:
        key = kwargs["key"]
        if key not in cache:
            cache[key] = {}
        cache = cache[key]

    alignment_is_float = math.isclose(alignment, round(alignment))

    if cache is not None and "rotation" in cache:
        rotation = transform = cache["rotation"].to(a.device, a.dtype)
    else:
        rotation = transform = orthogonal_procrustes(a_neurons, b_neurons, cancel_reflection=alignment_is_float)
        if cache is not None:
            cache["rotation"] = rotation.to("cpu", torch.float16)

    if alignment_is_float:
        transform = fractional_matrix_power(transform, alignment, cache)
    elif math.isclose(alignment, 0):
        transform = torch.eye(
            len(transform),
            dtype=transform.dtype,
            device=transform.device,
        )
    elif not math.isclose(alignment, 1):
        transform = torch.linalg.matrix_power(transform, round(alignment))

    if not torch.allclose(alpha, torch.zeros_like(alpha)):
        # interpolate the relationship between the neurons
        a_neurons = (1-alpha)*a_neurons + alpha*b_neurons@rotation.T

    a_neurons @= transform
    a_neurons += torch.lerp(a_centroid, b_centroid, alignment)
    return a_neurons.reshape_as(a)


@merge_method
def clamp(
    a: Parameter(Tensor),
    *bounds: Parameter(Tensor),
    stiffness: Parameter(float) = 0.0,
) -> Return(Tensor):
    maximums = functools.reduce(torch.maximum, bounds)
    minimums = functools.reduce(torch.minimum, bounds)

    if stiffness:
        bounds = torch.stack(bounds)
        average = bounds.mean(dim=0)

        smallest_positive = maximums
        largest_negative = minimums

        for i, bound in enumerate(bounds):
            smallest_positive = torch.where((smallest_positive >= bound) & (bound >= average), bound, smallest_positive)
            largest_negative = torch.where((largest_negative <= bound) & (bound <= average), bound, largest_negative)

        maximums = (1-stiffness)*maximums + stiffness*smallest_positive
        minimums = (1-stiffness)*minimums + stiffness*largest_negative

    return torch.minimum(torch.maximum(a, minimums), maximums)


@merge_method
def copy_distribution(
    a: Parameter(Tensor),
) -> Return(Tensor):
    return torch.randn_like(a) * a.std(correction=0) + a.mean()


# Special mode "TIES-STOCK" has been implemented by setting `apply_stock` > 0.0
# Special mode "TIES-GMEDIAN" has been implemented by setting `apply_median` > 0.0
@merge_method
def ties_sum_extended(  # aka add_difference_ties
    *models: Parameter(Tensor, "delta"),
    k: Parameter(float) = 0.2,
    vote_sgn: Parameter(bool) = False,
    apply_stock: Parameter(bool) = False,
    apply_median: Parameter(bool) = False,
    cos_eps: Parameter(float) = 1e-6,
    eps: Parameter(float) = 1e-6,
    maxiter: Parameter(int) = 100,
    ftol: Parameter(float) = 1e-20,
) -> Return(Tensor, "delta"):
    filtered_delta, param_counts = ties_sum_deltas(*models, k=k, vote_sgn=vote_sgn)

    if apply_median:
        # Model Stock
        t = 1.0 if apply_stock else get_model_stock_t(torch.unbind(filtered_delta), cos_eps=cos_eps)

        filtered_delta = filtered_delta.sum(dim=0)

        # $$ \tau_m $$
        return torch.nan_to_num(filtered_delta * t / param_counts)
    else:
        # $$ \tau_m $$, but in geometric median instead of arithmetic mean. Considered to replace model stock.
        filtered_delta = geometric_median.__wrapped__(*torch.unbind(filtered_delta), eps=eps, maxiter=maxiter, ftol=ftol)

        return torch.nan_to_num(filtered_delta)


# latex notes in reference to original implementation: https://arxiv.org/abs/2306.01708
# - `delta`: $$ \hat{\tau}_t $$
# - `signs`: $$ \gamma_t $$
# - `final_sign`: $$ \gamma_m^p = sgn(\Sigma_{t=1}^n \hat{\tau}_t^p) $$
# - `delta_filters`: $$ \{ \gamma_t^p = \gamma_m^p \} $$
# - `param_counts`: $$ |A^p| $$
# - `filtered_delta`: $$ \Sigma_{t\in{A^p}} \hat{\tau}_t^p $$
# - `return`: $$ \lambda * \tau_m $$
# Special mode "TIES-SOUP" has been implemented by setting `vote_sgn` > 0.0
# - `final_sign`: $$ \gamma_m^p = sgn(\Sigma_{t=1}^n \gamma_t^p) $$
@merge_method
def ties_sum(  # aka add_difference_ties
    *models: Parameter(Tensor, "delta"),
    k: Parameter(float) = 1.0,
    vote_sgn: Parameter(bool) = False,
) -> Return(Tensor, "delta"):
    filtered_delta, param_counts = ties_sum_deltas(*models, k=k, vote_sgn=vote_sgn)

    # $$ \tau_m $$
    return torch.nan_to_num(filtered_delta.sum(dim=0) / param_counts)


def ties_sum_deltas(
    *models: Tensor,
    k: float = 0.2,
    vote_sgn: bool = False,
):
    # Step 1: Trim redundant parameters

    # $$ \hat{\tau}_t $$ O(N) in space
    deltas = [
        # $$ keep_topk_reset_rest_to_zero(\tau_t, k) $$
        filter_top_k(m, k)
        for m in models
    ]
    deltas = torch.stack(deltas, dim=0)

    # Step 2: Elect Final Signs.

    # $$ \gamma_t $$
    signs = torch.sign(deltas)

    # $$ \gamma_m^p = sgn(\Sigma_{t=1}^n \hat{\tau}_t^p) $$ for normal TIES
    # $$ \gamma_m^p = sgn(\Sigma_{t=1}^n \gamma_t^p) $$ if "TIES-SOUP" is activated
    final_sign = torch.sign(torch.sum(deltas if vote_sgn else signs, dim=0))

    # Step 3: Disjoint merge.

    # $$ \{ \gamma_t^p = \gamma_m^p \} $$
    delta_filters = (signs == final_sign).float()

    # $$ |A^p| $$
    param_counts = torch.sum(delta_filters, dim=0)

    # $$ \Sigma_{t\in{A^P}} \hat{\tau}_t^p $$
    # (note that the sum is not performed here directly)
    filtered_delta = deltas * delta_filters

    return filtered_delta, param_counts


def filter_top_k(a: Tensor, k: float):
    k = max(int((1 - k) * a.numel()), 1)
    k_value, _ = a.flatten().abs().float().kthvalue(k)
    top_k_filter = (a.abs() >= k_value).float()
    return a * top_k_filter


@merge_method
def dropout(  # aka n-supermario
    *deltas: Parameter(Tensor, "delta"),
    probability: Parameter(float) = 0.9,
    rescale: Parameter(float) = 1.0,
    overlap: Parameter(float) = 1.0,
    overlap_emphasis: Parameter(float) = 0.0,
    seed: Parameter(int) = None,
) -> Return(Tensor, "delta"):
    if len(deltas) == 0:
        return 0

    delta0 = deltas[0]
    deltas = torch.stack(deltas)
    rng = np.random.default_rng(seed)

    if overlap % 2 == 1:
        masks = tuple(
            torch.from_numpy(rng.binomial(n=1, p=1 - probability, size=delta0.shape)).to(device=delta0.device, dtype=torch.bool)
            for _ in range(len(deltas))
        )
    else:
        ks = np.arange(2 ** len(deltas))
        pmf = overlapping_sets_pmf(len(deltas), probability, overlap, overlap_emphasis)
        masks = torch.from_numpy(rng.choice(ks, size=delta0.shape, p=pmf)).to(delta0.device)
        masks = torch.stack([masks & 2 ** i != 0 for i in range(len(deltas))])

    final_delta = torch.zeros_like(delta0)
    for mask, delta in zip(masks, deltas):
        final_delta[mask] += delta[mask]

    if probability == 1.0:
        rescalar = 1.0
    else:
        rescalar = (1.0 - probability) ** rescale
        rescalar = rescalar if math.isfinite(rescalar) else 1
    return final_delta / masks.sum(0).clamp(1) / rescalar


def overlapping_sets_pmf(n, p, overlap: float, overlap_emphasis):
    if np.isclose(overlap, round(overlap)):
        if round(overlap) % 2 == 0:
            pmf = np.array([1/n*float(bin(i).count("1") == 1) for i in range(1, 2**n)])
        else:
            pmf = np.array([0 for _ in range(1, 2**n - 1)] + [1])
    else:
        if math.floor(overlap) % 2 == 1:
            overlap = -overlap

        tan_overlap = np.tan(np.pi * (overlap - 0.5))
        pmf = np.zeros(2 ** n - 1)
        for i in range(1, 2 ** n):
            num_sets = bin(i).count("1")
            pmf[i-1] = tan_overlap*(num_sets - n/2)
        pmf = np.exp(pmf) / np.sum(np.exp(pmf))

    binomial_pmf = binom.pmf(np.arange(1, n + 1), n, p)
    expanded_binomial_pmf = np.zeros(2 ** n - 1)
    for i in range(1, 2 ** n):
        num_sets = bin(i).count("1")
        expanded_binomial_pmf[i-1] = binomial_pmf[num_sets-1] / binomial_coefficient_np(n, num_sets)
    expanded_binomial_pmf /= expanded_binomial_pmf.sum()

    pmf = torch.lerp(
        pmf,
        torch.lerp(pmf, expanded_binomial_pmf, 1-abs(2*overlap-1)),
        overlap_emphasis,
    )
    return np.concatenate([[p], pmf * (1 - p)])


# Part of TIES w/ DARE
# Hyperparameters defauled to values proposed to paper.
# Special mode "DROP" has been implemented by setting `no_rescale` > 0.0
# - `return`: $$ \hat{\delta}^t = \tilde{\delta}^t $$
@merge_method
def ties_sum_with_dropout(
    *deltas: Parameter(Tensor, "delta"),
    probability: Parameter(float) = 0.9,
    no_rescale: Parameter(bool) = False,
    k: Parameter(float) = 0.2,
    vote_sgn: Parameter(bool) = False,
    apply_stock: Parameter(bool) = False,
    cos_eps: Parameter(float) = 1e-6,
    apply_median: Parameter(bool) = False,
    eps: Parameter(float) = 1e-6,
    maxiter: Parameter(int) = 100,
    ftol: Parameter(float) = 1e-20,
    seed: Parameter(int) = None,
) -> Return(Tensor, "delta"):
    # Set seed
    generator = torch.Generator(device=deltas[0].device)
    if seed is not None:
        generator.manual_seed(seed)

    # Under "Dropout", delta will be 0 by definition. Multiply it (Hadamard product) will return 0 also.
    # $$ \tilde{\delta}^t = (1 - m^t) \odot \delta^t $$
    deltas = tuple(
        delta * torch.bernoulli(torch.full(delta.shape, 1 - probability, device=deltas[0].device, dtype=deltas[0].dtype), generator=generator)
        for delta in deltas
    )

    # $$ \tilde{\delta}^t = \tau_m = \hat{\tau}_t $$ O(N) in space
    deltas = ties_sum_extended.__wrapped__(
        *deltas,
        k=k,
        vote_sgn=vote_sgn,
        apply_stock=apply_stock,
        cos_eps=cos_eps,
        apply_median=apply_median,
        eps=eps,
        maxiter=maxiter,
        ftol=ftol,
    )

    if math.isclose(probability, 1.0):
        # Corner case
        return deltas * 0.0
    elif no_rescale:
        # Rescale
        # $$ \hat{\delta}^t = \tilde{\delta}^t / (1-p) $$
        return deltas / (1.0 - probability) 
    else:
        # No rescale
        # $$ \hat{\delta}^t = \tilde{\delta}^t $$
        return deltas


def binomial_coefficient_np(n, k):
    if k > n - k:
        k = n - k
    result = np.int64(1)
    for i in range(1, k+1):
        result = result * (n - i + 1) // i
    return result


# src: https://github.com/arcee-ai/mergekit/blob/main/mergekit/merge_methods/model_stock.py
@merge_method
def model_stock(
    *deltas: Parameter(Tensor, "delta"),
    cos_eps: Parameter(float) = 1e-6,
) -> Return(Tensor, "delta"):
    w_avg = n_average.__wrapped__(*deltas)

    # t can get inf so handle with care
    t = get_model_stock_t(deltas, cos_eps)

    # return w_h. Notice that w_0 is 0 here.
    return torch.nan_to_num(t * w_avg)


# The guess from mergekit: Average of cos(theta). Expected value is 0, somehow match with paper.
# However this may be very unstable, and the range is still -1 to 1.
def get_model_stock_t(deltas: Sequence[Tensor], cos_eps: float):
    n = len(deltas)

    # Generator function
    cos = torch.nn.CosineSimilarity(dim=-1, eps=cos_eps)
    cos_thetas = [cos(deltas[i], deltas[i + 1]) for i, _ in enumerate(deltas) if (i + 1) < n]

    # Still a vector
    cos_theta = torch.stack(cos_thetas).mean(dim=0)

    # Convert to column vector for multiplication
    t = (n * cos_theta / (1 + (n - 1) * cos_theta)).unsqueeze(-1)
    return t


# src: https://github.com/krishnap25/geom_median/blob/main/src/geom_median/torch/weiszfeld_list_of_array.py
@merge_method
def geometric_median(
    *models: Parameter(Tensor),
    eps: Parameter(float) = 1e-6,
    maxiter: Parameter(int) = 100,
    ftol: Parameter(float) = 1e-20,
) -> Return(Tensor):
    weights = torch.ones(len(models), device=models[0].device)

    # initialize median estimate at mean
    median = weighted_average_tuple(models, weights)
    new_weights = weights
    objective_value = geometric_median_objective(median, models, weights)

    # Weiszfeld iterations
    for _ in range(max(0, maxiter)):
        prev_obj_value = objective_value
        denom = torch.stack([torch.dist(p, median) for p in models])
        new_weights = weights / torch.clamp(denom, min=eps)
        median = weighted_average_tuple(models, new_weights)

        objective_value = geometric_median_objective(median, models, weights)
        if abs(prev_obj_value - objective_value) <= ftol * objective_value:
            break

    return weighted_average_tuple(models, new_weights)


def weighted_average_tuple(points: Tuple, weights):
    return torch.sum(torch.stack([p * weights[i] for i, p in enumerate(points)]), dim=0) / weights.sum()


def geometric_median_objective(median, points: Tuple, weights):
    return torch.mean(torch.stack([torch.dist(point, median) for point in points]) * weights)


@merge_method
def truncate_rank(
    a: Parameter(Tensor, merge_space="delta"),
    rank_ratio: Parameter(float) = 0.5,
) -> Return(Tensor, merge_space="delta"):
    if a.dim() < 2:
        return a

    original_shape = a.shape
    a_2d = a.flatten(start_dim=1)
    max_rank = min(a_2d.shape)
    target_rank = min(max(round(max_rank * rank_ratio), 0), max_rank)
    if target_rank == max_rank:
        return a
    if target_rank == 0:
        return torch.zeros_like(a)

    svd_driver = "gesvda" if a.is_cuda else None
    u, s, vt = torch_svd_lowrank(a_2d, q=target_rank, full_matrices=False, driver=svd_driver)
    return (u @ torch.diag(s) @ vt).reshape(original_shape)


@merge_method
def fallback(
    a: Parameter(StateDict[T]),
    default: Parameter(StateDict[T]),
    **kwargs,
) -> Return(T):
    key = kwargs["key"]
    try:
        return a[key]
    except StateDictKeyError:
        return default[key]


@merge_method
def cast(
    a: Parameter(Tensor),
    device: Parameter(str) = None,
    dtype: Parameter(str) = None,
) -> Return(Tensor):
    to_kwargs = {}
    if device is not None:
        to_kwargs["device"] = device

    if dtype is not None:
        if dtype not in cast_dtype_map:
            raise ValueError(f"Unknown dtype {dtype}. Possible values are None, {', '.join(cast_dtype_map.keys())}")
        to_kwargs["dtype"] = cast_dtype_map[dtype]

    return a.to(**to_kwargs)


cast_dtype_map = {
    "float64": torch.float64,
    "int64": torch.int64,
    "float32": torch.float32,
    "int32": torch.int32,
    "float16": torch.float16,
    "bfloat16": torch.bfloat16,
    "int16": torch.int16,
    "float8_e4m3fn": torch.float8_e4m3fn,
    "float8_e5m2": torch.float8_e5m2,
    "int8": torch.int8,
    "bool": torch.bool,
}
for dtype_str in ("uint8", "uint16", "uint32", "uint64"):
    if hasattr(torch, dtype_str):
        cast_dtype_map[dtype_str] = getattr(torch, dtype_str)
cast_dtype_map_reversed = {v: k for k, v in cast_dtype_map.items()}


@merge_method
def get_dtype(
    a: Parameter(Tensor),
) -> Return(str, "param"):
    return cast_dtype_map_reversed[a.dtype]


@merge_method
def get_device(
    a: Parameter(Tensor),
) -> Return(str, "param"):
    return str(a.device)


@merge_method
def pick_component(
    a: Parameter(StateDict[T]),
    component: Parameter(str, "param"),
    **kwargs,
) -> Return(T):
    if component not in a.model_config.components():
        raise ValueError(
            f'Component "{component}" does not exist in config "{a.model_config.identifier}". '
            f"Valid components: {tuple(a.model_config.components())}"
        )

    key = kwargs["key"]
    if key in a.model_config.components()[component].keys():
        return a[key]
    else:
        raise StateDictKeyError(key)


@merge_method
def omit_component(
    a: Parameter(StateDict[T]),
    component: Parameter(str, "param"),
    **kwargs,
) -> Return(T):
    if component not in a.model_config.components():
        raise ValueError(
            f'Component "{component}" does not exist in config "{a.model_config.identifier}". '
            f"Valid components: {tuple(a.model_config.components())}"
        )

    key = kwargs["key"]
    if key in a.model_config.components()[component].keys():
        raise StateDictKeyError(key)
    else:
        return a[key]



import abc
import dataclasses
import inspect
import pathlib
import re
import fuzzywuzzy.process
import torch
import yaml
from collections import OrderedDict
from sd_mecha.streaming import TensorMetadata
from typing import Dict, List, Iterable, Mapping, Protocol, runtime_checkable, Optional
try:
    from yaml import CLoader as YamlLoader
except ImportError:
    from yaml import Loader as YamlLoader


StateDictKey = str


@dataclasses.dataclass
class KeyMetadata:
    shape: Optional[List[int]] | torch.Size
    dtype: Optional[str] | torch.dtype
    aliases: Iterable[str] = dataclasses.field(default_factory=tuple, metadata={"exclude": lambda p: bool(p)})
    optional: bool = dataclasses.field(default=False, metadata={"exclude": lambda p: not p})

    def __post_init__(self):
        if isinstance(self.shape, list):
            self.shape = torch.Size(self.shape)
        if isinstance(self.dtype, str):
            self.dtype = getattr(torch, self.dtype)

    def metadata(self) -> TensorMetadata:
        return TensorMetadata(self.shape, self.dtype)


@dataclasses.dataclass
class ModelComponent:
    _keys: Mapping[StateDictKey, KeyMetadata] = dataclasses.field(metadata={"serial_name": "keys"})

    def __post_init__(self):
        keys = OrderedDict()
        for k, v in self._keys.items():
            if isinstance(v, Mapping):
                keys[k] = KeyMetadata(**v)
            else:
                keys[k] = v
        self._keys = keys

    def keys(self) -> Mapping[StateDictKey, KeyMetadata]:
        return OrderedDict(
            (k, v)
            for k, v in self._keys.items()
        )

    def metadata(self) -> Mapping[StateDictKey, TensorMetadata]:
        return OrderedDict(
            (k, v.metadata())
            for k, v in self._keys.items()
        )

    def aliases(self) -> Mapping[StateDictKey, Iterable[StateDictKey]]:
        return OrderedDict(
            (k, v.aliases)
            for k, v in self._keys.items()
        )


@runtime_checkable
class ModelConfig(Protocol):
    def eq(self, other):
        other_identifier = getattr(other, "identifier", None)
        return self.identifier == other_identifier

    def __repr__(self):
        return f"<model config '{self.identifier}'>"

    @property
    @abc.abstractmethod
    def identifier(self) -> str:
        ...

    @abc.abstractmethod
    def get_architecture_identifier(self) -> str:
        ...

    @abc.abstractmethod
    def get_implementation_identifier(self) -> str:
        ...

    @abc.abstractmethod
    def components(self) -> Mapping[str, ModelComponent]:
        ...

    @abc.abstractmethod
    def keys(self) -> Mapping[StateDictKey, KeyMetadata]:
        ...

    @abc.abstractmethod
    def metadata(self) -> Mapping[StateDictKey, TensorMetadata]:
        ...

    @abc.abstractmethod
    def aliases(self) -> Mapping[StateDictKey, Iterable[StateDictKey]]:
        ...


@dataclasses.dataclass
class ModelConfigImpl(ModelConfig):
    _identifier: str = dataclasses.field(metadata={"serial_name": "identifier"})
    _components: Mapping[str, ModelComponent] = dataclasses.field(metadata={"serial_name": "components"})

    _keys_cache: Mapping[StateDictKey, KeyMetadata] = dataclasses.field(default=None, init=False, hash=False, compare=False, metadata={"exclude": True})
    _metadata_cache: Mapping[StateDictKey, TensorMetadata] = dataclasses.field(default=None, init=False, hash=False, compare=False, metadata={"exclude": True})
    _aliases_cache: Mapping[StateDictKey, Iterable[StateDictKey]] = dataclasses.field(default=None, init=False, hash=False, compare=False, metadata={"exclude": True})

    def __post_init__(self):
        if not re.fullmatch("[a-z0-9._+]+-[a-z0-9._+]+", self.identifier):
            raise ValueError(
                f"Identifier of model {self.identifier} is invalid: "
                "it must only contain lowercase alphanumerical characters, '.', '_' or '+', "
                "and must match the pattern '<architecture>-<implementation>'. "
                "An example of valid identifier is 'flux_dev-flux'"
            )

        components = OrderedDict()
        for k, v in self._components.items():
            if isinstance(v, Mapping):
                components[k] = ModelComponent(v)
            else:
                components[k] = v
        self._components = components

    @property
    def identifier(self) -> str:
        return self._identifier

    def get_architecture_identifier(self):
        return self.identifier.split("-")[0]

    def get_implementation_identifier(self):
        return self.identifier.split("-")[1]

    def components(self) -> Mapping[str, ModelComponent]:
        return self._components

    def keys(self) -> Mapping[StateDictKey, KeyMetadata]:
        if self._keys_cache is None:
            self._keys_cache = OrderedDict(
                (k, v)
                for component in self.components().values()
                for k, v in component.keys().items()
            )
        return self._keys_cache

    def metadata(self) -> Mapping[StateDictKey, TensorMetadata]:
        if self._metadata_cache is None:
            self._metadata_cache = OrderedDict(
                (k, v)
                for component in self.components().values()
                for k, v in component.metadata().items()
            )
        return self._metadata_cache

    def aliases(self) -> Mapping[StateDictKey, Iterable[StateDictKey]]:
        if self._aliases_cache is None:
            self._aliases_cache = OrderedDict(
                (k, v)
                for component in self.components().values()
                for k, v in component.aliases().items()
            )
        return self._aliases_cache


def ModelConfigImpl__init__patch(self, *args, **kwargs):
    for field in dataclasses.fields(ModelConfigImpl):
        if "serial_name" in field.metadata and field.metadata["serial_name"] in kwargs:
            kwargs[field.name] = kwargs.pop(field.metadata["serial_name"])

    ModelConfigImpl__init__(self, *args, **kwargs)


ModelConfigImpl__init__ = ModelConfigImpl.__init__
ModelConfigImpl.__init__ = ModelConfigImpl__init__patch


class LazyModelConfigBase(ModelConfig):
    def __init__(self):
        self.underlying_config = None

    @classmethod
    def __init_subclass__(cls):
        super().__init_subclass__()
        for name, value in inspect.getmembers(ModelConfig):
            if (
                (inspect.isfunction(value) or isinstance(value, property)) and
                getattr(value, "__isabstractmethod__", False) and
                name not in cls.__dict__
            ):
                # implement all remaining abstract methods as delegating to underlying_config
                setattr(cls, name, property(lambda self, name=name: resolve_lazy_model_config_attribute(self, name=name)))

    def _ensure_config(self) -> None:
        if self.underlying_config is not None:
            return

        self.underlying_config = self.create_config()

    @abc.abstractmethod
    def create_config(self) -> ModelConfig:
        raise NotImplementedError


def resolve_lazy_model_config_attribute(self: LazyModelConfigBase, name: str):
    self._ensure_config()
    attribute = getattr(self.underlying_config, name)
    if inspect.ismethod(attribute):
        method = attribute.__func__.__get__(self.underlying_config, self.underlying_config.__class__)
        return method
    return attribute


class StructuralModelConfig(ModelConfig):
    def __init__(self, keys: Mapping[StateDictKey, KeyMetadata]):
        self._keys_cache = keys
        self._metadata_cache = None
        self._aliases_cache = None

    @property
    def identifier(self) -> str:
        return "structural"

    def get_architecture_identifier(self) -> str:
        return self.identifier

    def get_implementation_identifier(self) -> str:
        return ""

    def components(self) -> Mapping[str, ModelComponent]:
        return {"keys": ModelComponent(self.keys())}

    def keys(self) -> Mapping[StateDictKey, KeyMetadata]:
        return self._keys_cache

    def metadata(self) -> Mapping[StateDictKey, TensorMetadata]:
        if self._metadata_cache is None:
            self._metadata_cache = OrderedDict(
                (k, v.metadata())
                for k, v in self.keys().items()
            )
        return self._metadata_cache

    def aliases(self) -> Mapping[StateDictKey, Iterable[StateDictKey]]:
        if self._aliases_cache is None:
            self._aliases_cache = OrderedDict(
                (k, v.aliases)
                for k, v in self.keys().items()
            )
        return self._aliases_cache


class InferModelConfig(ModelConfig):
    def eq(self, other):
        raise RuntimeError("the config has not yet been inferred")

    @property
    def identifier(self) -> str:
        return "infer"

    def get_architecture_identifier(self) -> str:
        return self.identifier

    def get_implementation_identifier(self) -> str:
        return ""

    def components(self) -> Mapping[str, ModelComponent]:
        raise RuntimeError("the config has not yet been inferred")

    def keys(self) -> Mapping[StateDictKey, KeyMetadata]:
        raise RuntimeError("the config has not yet been inferred")

    def metadata(self) -> Mapping[StateDictKey, TensorMetadata]:
        raise RuntimeError("the config has not yet been inferred")

    def aliases(self) -> Mapping[StateDictKey, Iterable[StateDictKey]]:
        raise RuntimeError("the config has not yet been inferred")


INFER = InferModelConfig()


class YamlModelConfig(LazyModelConfigBase):
    def __init__(self, yaml_config_file: pathlib.Path):
        super().__init__()
        self.yaml_config_file = yaml_config_file
        self._identifier = yaml_config_file.stem

    @property
    def identifier(self) -> str:
        return self._identifier

    def create_config(self) -> ModelConfig:
        with open(self.yaml_config_file, "r") as f:
            yaml_config = f.read()

        return from_yaml(yaml_config)


_model_configs_registry_base: Dict[str, ModelConfig] = {}
_model_configs_registry_aux: Dict[str, ModelConfig] = {}


def serialize(obj):
    if isinstance(obj, ModelComponent):
        return serialize(obj.keys)
    elif dataclasses.is_dataclass(obj):
        return {
            field.metadata.get("serial_name", field.name): serialize(getattr(obj, field.name))
            for field in dataclasses.fields(obj)
            if not (
                callable(field.metadata.get("exclude")) and field.metadata["exclude"](getattr(obj, field.name))
                or field.metadata.get("exclude", False)
            )
        }
    elif isinstance(obj, (str, int, float, type(None))):
        return obj
    elif isinstance(obj, torch.dtype):
        return str(obj).split(".")[1]
    elif isinstance(obj, torch.Size):
        return list(obj)
    elif isinstance(obj, Mapping):
        return {str(k): serialize(v) for k, v in obj.items()}
    elif isinstance(obj, Iterable) and not isinstance(obj, bytes):
        return [serialize(v) for v in obj]
    else:
        raise ValueError(f"Cannot serialize {obj!r}")


def to_yaml(model_config: ModelConfig) -> str:
    dict_config = serialize(model_config)
    old_representers = yaml.SafeDumper.yaml_representers.copy()

    def flow_style_list_representer(dumper, data):
        return dumper.represent_sequence('tag:yaml.org,2002:seq', data, flow_style=True)

    def flow_style_dict_representer(dumper, data):
        if "shape" in data and "dtype" in data:
            return dumper.represent_mapping('tag:yaml.org,2002:map', data, flow_style=True)
        else:
            return dumper.represent_mapping('tag:yaml.org,2002:map', data)

    try:
        yaml.SafeDumper.add_representer(list, flow_style_list_representer)
        yaml.SafeDumper.add_representer(dict, flow_style_dict_representer)
        return yaml.safe_dump(dict_config, width=2**64, sort_keys=False)
    finally:
        yaml.SafeDumper.yaml_representers.clear()
        yaml.SafeDumper.yaml_representers.update(old_representers)


def from_yaml(yaml_config: str) -> ModelConfig:
    dict_config = yaml.load(yaml_config, Loader=YamlLoader)
    return ModelConfigImpl(**dict_config)


def register(config: ModelConfig):
    if config.identifier in _model_configs_registry_base or config.identifier in _model_configs_registry_aux:
        raise ValueError(f"Model {config.identifier} already exists")

    _model_configs_registry_base[config.identifier] = config


def register_aux(config: ModelConfig):
    if config.identifier in _model_configs_registry_base or config.identifier in _model_configs_registry_aux:
        raise ValueError(f"Model {config.identifier} already exists")

    _model_configs_registry_aux[config.identifier] = config


def resolve(identifier: str) -> ModelConfig:
    try:
        return _model_configs_registry_base[identifier]
    except KeyError:
        pass

    try:
        return _model_configs_registry_aux[identifier]
    except KeyError:
        pass

    if identifier == INFER.identifier:
        return INFER

    if identifier == "structural":
        raise ValueError(
            "the 'structural' model config is not a unique object, "
            "it needs to be manually instantiated in this way: model_configs.StructuralModelConfig(...)"
        )

    suggestions = fuzzywuzzy.process.extractOne(identifier, _model_configs_registry_base.keys())
    postfix = ""
    if suggestions is not None:
        postfix = f". Nearest match is '{suggestions[0]}'"
    raise ValueError(f"unknown model implementation: {identifier}{postfix}")


def get_all() -> List[ModelConfig]:
    res = get_all_base() + get_all_aux()
    res.sort(key=lambda c: c.identifier, reverse=True)
    return res


def get_all_base() -> List[ModelConfig]:
    return list(_model_configs_registry_base.values())


def get_all_aux() -> List[ModelConfig]:
    return list(_model_configs_registry_aux.values())



import functools
from typing import List, Dict, Set, Tuple
import fuzzywuzzy.process


class MergeSpace:
    def __init__(self, identifier: str):
        self.identifier = identifier

    def __eq__(self, other):
        if isinstance(other, str):
            other = MergeSpace(other)
        return self.identifier == other.identifier

    def __hash__(self):
        return hash(self.identifier)

    def __repr__(self):
        return f"MergeSpace('{self.identifier}')"


class MergeSpaceSymbol:
    def __init__(self, *merge_spaces: Tuple[str | MergeSpace, ...]):
        self.merge_spaces = {
            MergeSpace(merge_space) if isinstance(merge_space, str) else merge_space
            for merge_space in merge_spaces
        }


AnyMergeSpace = Set[MergeSpace] | MergeSpaceSymbol


def register_merge_space(identifier: str):
    if identifier in _merge_space_registry:
        raise ValueError(f"merge space {identifier} already exists")
    _merge_space_registry[identifier] = MergeSpace(identifier)


def resolve(identifier: str) -> MergeSpace:
    try:
        return _merge_space_registry[identifier]
    except KeyError as e:
        suggestion = fuzzywuzzy.process.extractOne(str(e), _merge_space_registry.keys())[0]
        raise ValueError(f"unknown merge space: {e}. Nearest match is '{suggestion}'")


def get_identifiers(merge_space: AnyMergeSpace) -> List[str]:
    if isinstance(merge_space, Set):
        return [merge_space.identifier for merge_space in merge_space]
    elif isinstance(merge_space, MergeSpaceSymbol):
        return get_identifiers(merge_space.merge_spaces)
    else:
        raise TypeError(f"expected {MergeSpaceSymbol.__name__} or Tuple[{MergeSpace.__name__}, ...], got {type(merge_space)}")


def get_all() -> Set[MergeSpace]:
    return set(_merge_space_registry.values())


@functools.cache
def _register_builtin_merge_spaces():
    global _builtin_merge_spaces
    for builtin_merge_space in _builtin_merge_spaces:
        register_merge_space(builtin_merge_space)


_merge_space_registry: Dict[str, MergeSpace] = {}
_builtin_merge_spaces = [
    "weight",
    "delta",
    "param",
]
_register_builtin_merge_spaces()



from sd_mecha import merge_method, Parameter, Return, StateDict, StateDictKeyError
from sd_mecha.extensions import model_configs
from torch import Tensor


# hf to sd conversion src:
# https://github.com/huggingface/diffusers/blob/main/scripts/convert_diffusers_to_original_stable_diffusion.py


sdxl_diffusers_unet_config = model_configs.resolve('sdxl-diffusers_unet_only')
sdxl_sgm_config = model_configs.resolve('sdxl-sgm')


@merge_method(
    identifier=f"convert_'{sdxl_diffusers_unet_config.identifier}'_to_'{sdxl_sgm_config.identifier}'",
    is_conversion=True,
)
def convert_sdxl_diffusers_unet_to_original(
    diffusers_sd: Parameter(StateDict[Tensor], model_config=sdxl_diffusers_unet_config),
    **kwargs,
) -> Return(Tensor, model_config=sdxl_sgm_config):
    sgm_key = kwargs["key"]
    if not sgm_key.startswith("model.diffusion_model"):
        raise StateDictKeyError(sgm_key)

    kohya_key = ".".join(sgm_key.split(".")[2:])  # model.diffusion_model.

    for sd_part, hf_part in unet_conversion_map_layer.items():
        kohya_key = kohya_key.replace(sd_part, hf_part)

    if "resnets" in kohya_key:
        for sd_part, hf_part in unet_conversion_map_resnet.items():
            kohya_key = kohya_key.replace(sd_part, hf_part)

    kohya_key = unet_conversion_map.get(kohya_key, kohya_key)
    return diffusers_sd[kohya_key]


unet_conversion_map = {
    # (original: huggingface)
    "time_embed.0.weight": "time_embedding.linear_1.weight",
    "time_embed.0.bias": "time_embedding.linear_1.bias",
    "time_embed.2.weight": "time_embedding.linear_2.weight",
    "time_embed.2.bias": "time_embedding.linear_2.bias",
    "input_blocks.0.0.weight": "conv_in.weight",
    "input_blocks.0.0.bias": "conv_in.bias",
    "out.0.weight": "conv_norm_out.weight",
    "out.0.bias": "conv_norm_out.bias",
    "out.2.weight": "conv_out.weight",
    "out.2.bias": "conv_out.bias",
    # the following are for sdxl
    "label_emb.0.0.weight": "add_embedding.linear_1.weight",
    "label_emb.0.0.bias": "add_embedding.linear_1.bias",
    "label_emb.0.2.weight": "add_embedding.linear_2.weight",
    "label_emb.0.2.bias": "add_embedding.linear_2.bias",
}

unet_conversion_map_resnet = {
    # (original: huggingface)
    "in_layers.0": "norm1",
    "in_layers.2": "conv1",
    "out_layers.0": "norm2",
    "out_layers.3": "conv2",
    "emb_layers.1": "time_emb_proj",
    "skip_connection": "conv_shortcut",
}


unet_conversion_map_layer = {}
# hardcoded number of downblocks and resnets/attentions...
# would need smarter logic for other networks.
for i in range(3):
    # loop over downblocks/upblocks

    for j in range(2):
        # loop over resnets/attentions for downblocks
        hf_down_res_prefix = f"down_blocks.{i}.resnets.{j}."
        sd_down_res_prefix = f"input_blocks.{3*i + j + 1}.0."
        unet_conversion_map_layer[sd_down_res_prefix] = hf_down_res_prefix

        if i > 0:
            hf_down_atn_prefix = f"down_blocks.{i}.attentions.{j}."
            sd_down_atn_prefix = f"input_blocks.{3*i + j + 1}.1."
            unet_conversion_map_layer[sd_down_atn_prefix] = hf_down_atn_prefix

    for j in range(4):
        # loop over resnets/attentions for upblocks
        hf_up_res_prefix = f"up_blocks.{i}.resnets.{j}."
        sd_up_res_prefix = f"output_blocks.{3*i + j}.0."
        unet_conversion_map_layer[sd_up_res_prefix] = hf_up_res_prefix

        if i < 2:
            # no attention layers in up_blocks.0
            hf_up_atn_prefix = f"up_blocks.{i}.attentions.{j}."
            sd_up_atn_prefix = f"output_blocks.{3 * i + j}.1."
            unet_conversion_map_layer[sd_up_atn_prefix] = hf_up_atn_prefix

    if i < 3:
        # no downsample in down_blocks.3
        hf_downsample_prefix = f"down_blocks.{i}.downsamplers.0.conv."
        sd_downsample_prefix = f"input_blocks.{3*(i+1)}.0.op."
        unet_conversion_map_layer[sd_downsample_prefix] = hf_downsample_prefix

        # no upsample in up_blocks.3
        hf_upsample_prefix = f"up_blocks.{i}.upsamplers.0."
        sd_upsample_prefix = f"output_blocks.{3*i + 2}.{1 if i == 0 else 2}."
        unet_conversion_map_layer[sd_upsample_prefix] = hf_upsample_prefix
unet_conversion_map_layer["output_blocks.2.2.conv."] = "output_blocks.2.1.conv."

hf_mid_atn_prefix = "mid_block.attentions.0."
sd_mid_atn_prefix = "middle_block.1."
unet_conversion_map_layer[sd_mid_atn_prefix] = hf_mid_atn_prefix
for j in range(2):
    hf_mid_res_prefix = f"mid_block.resnets.{j}."
    sd_mid_res_prefix = f"middle_block.{2*j}."
    unet_conversion_map_layer[sd_mid_res_prefix] = hf_mid_res_prefix



import torch
from torch import Tensor
from typing import Optional, Tuple, Dict


def orthogonal_procrustes(a, b, cancel_reflection: bool = False):
    if a.shape != b.shape:
        raise ValueError(f"a {tuple(a.shape)} and b {tuple(b.shape)} must have the same shape")

    if not cancel_reflection and a.shape[0] + 10 < a.shape[1]:
        svd_driver = "gesvdj" if a.is_cuda else None
        u, _, v_t = torch_svd_lowrank(a.T @ b, q=a.shape[0] + 10, driver=svd_driver, full_matrices=cancel_reflection)
    else:
        svd_driver = "gesvd" if a.is_cuda else None
        u, _, v_t = torch.linalg.svd(a.T @ b, driver=svd_driver)
        if cancel_reflection:
            u[:, -1] /= torch.slogdet(u)[0] * torch.slogdet(v_t)[0]

    transform = u @ v_t
    return transform


def fractional_matrix_power(matrix: Tensor, power: float, cache: Optional[Dict[str, Tensor]] = None):
    if cache is not None and "eigenvalues" in cache:
        complex_dtype = torch_complex_dtype_map[matrix.dtype]
        eigenvalues = cache["eigenvalues"].to(matrix.device, complex_dtype)
        eigenvectors = cache["eigenvectors"].to(matrix.device, complex_dtype)
        eigenvectors_inv = cache["eigenvectors_inv"].to(matrix.device, complex_dtype)
    else:
        eigenvalues, eigenvectors = torch.linalg.eig(matrix)
        eigenvectors_inv = torch.linalg.inv(eigenvectors)
        if cache is not None:
            cache["eigenvalues"] = eigenvalues.to("cpu", torch.complex32)
            cache["eigenvectors"] = eigenvectors.to("cpu", torch.complex32)
            cache["eigenvectors_inv"] = eigenvectors_inv.to("cpu", torch.complex32)

    eigenvalues.pow_(power)
    result = eigenvectors @ torch.diag(eigenvalues) @ eigenvectors_inv
    return result.real.to(dtype=matrix.dtype)


torch_complex_dtype_map = {
    torch.bfloat16: torch.complex32,
    torch.float16: torch.complex32,
    torch.float32: torch.complex64,
    torch.float64: torch.complex128,
}


# need to redefine torch.svd_lowrank to specify the svd driver and for full_matrices support
def torch_svd_lowrank(
    A: Tensor,
    q: Optional[int] = 6,
    niter: Optional[int] = 2,
    driver: Optional[str] = None,
    full_matrices: Optional[bool] = True,
) -> Tuple[Tensor, Tensor, Tensor]:
    q = 6 if q is None else q
    m, n = A.shape[-2:]
    A_t = A.T

    # Algorithm 5.1 in Halko et al 2009, slightly modified to reduce
    # the number conjugate and transpose operations
    assert m < n or n > q
    # computing the SVD approximation of a transpose in
    # order to keep B shape minimal (the m < n case) or the V
    # shape small (the n > q case)
    Q = get_approximate_basis(A_t, q, niter=niter)
    Q_c = Q.conj()
    B_t = A @ Q_c
    assert B_t.shape[-2] == m, (B_t.shape, m)
    assert B_t.shape[-1] == q, (B_t.shape, q)
    assert B_t.shape[-1] <= B_t.shape[-2], B_t.shape
    U, S, Vh = torch.linalg.svd(B_t, driver=driver, full_matrices=full_matrices)
    V = Q @ Vh.mH
    if full_matrices:
        V = orthogonal_extend(V)

    return U, S, V.mH


def get_approximate_basis(A: Tensor, q: int, niter: Optional[int] = 2) -> Tensor:
    niter = 2 if niter is None else niter
    m, n = A.shape[-2:]
    dtype = get_floating_dtype(A)

    R = torch.randn(n, q, dtype=dtype, device=A.device)

    A_H = A.mH
    Q = torch.linalg.qr(A @ R).Q
    for i in range(niter):
        Q = torch.linalg.qr(A_H @ Q).Q
        Q = torch.linalg.qr(A @ Q).Q

    return Q


def orthogonal_extend(A: torch.Tensor) -> torch.Tensor:
    m, n = A.shape
    if m <= n:
        return A

    proj = torch.eye(m, device=A.device, dtype=A.dtype) - A @ A.mH
    proj @= torch.randn(m, m - n, device=A.device, dtype=A.dtype)
    A_extension = torch.linalg.householder_product(*torch.geqrf(proj))
    return torch.cat((A, A_extension), dim=1)


def get_floating_dtype(A):
    dtype = A.dtype
    if dtype.is_floating_point:
        return dtype
    return torch.float32
